from hatch.index.errors import ArtifactMetadataError

MULTIPLE_USE_METADATA_FIELDS = {
    'classifier',
    'dynamic',
    'license_file',
    'obsoletes_dist',
    'platform',
    'project_url',
    'provides_dist',
    'provides_extra',
    'requires_dist',
    'requires_external',
    'supported_platform',
}
RENAMED_METADATA_FIELDS = {'classifier': 'classifiers', 'project_url': 'project_urls'}


def get_wheel_form_data(artifact):
    import zipfile

    from packaging.tags import parse_tag

    with zipfile.ZipFile(str(artifact), 'r') as zip_archive:
        for path in zip_archive.namelist():
            root = path.split('/', 1)[0]
            if root.endswith('.dist-info'):
                dist_info_dir = root
                break
        else:  # no cov
            message = f'Could not find the `.dist-info` directory in wheel: {artifact}'
            raise ArtifactMetadataError(message)

        try:
            with zip_archive.open(f'{dist_info_dir}/METADATA') as zip_file:
                metadata_file_contents = zip_file.read().decode('utf-8')
        except KeyError:  # no cov
            message = f'Could not find a `METADATA` file in the `{dist_info_dir}` directory'
            raise ArtifactMetadataError(message) from None
        else:
            data = parse_headers(metadata_file_contents)

    data['filetype'] = 'bdist_wheel'

    # Examples:
    # cryptography-3.4.7-pp37-pypy37_pp73-manylinux2014_x86_64.whl -> pp37
    # hatchling-1rc1-py2.py3-none-any.whl -> py2.py3
    tag_component = '-'.join(artifact.stem.split('-')[-3:])
    data['pyversion'] = '.'.join(sorted({tag.interpreter for tag in parse_tag(tag_component)}))

    return data


def get_sdist_form_data(artifact):
    import tarfile

    with tarfile.open(str(artifact), 'r:gz') as tar_archive:
        pkg_info_dir_parts = []
        for tar_info in tar_archive:
            if tar_info.isfile():
                pkg_info_dir_parts.append(tar_info.name.split('/', 1)[0])
                break
        else:  # no cov
            message = f'Could not find any files in sdist: {artifact}'
            raise ArtifactMetadataError(message)

        pkg_info_dir_parts.append('PKG-INFO')
        pkg_info_path = '/'.join(pkg_info_dir_parts)
        try:
            with tar_archive.extractfile(pkg_info_path) as tar_file:
                metadata_file_contents = tar_file.read().decode('utf-8')
        except KeyError:  # no cov
            message = f'Could not find file: {pkg_info_path}'
            raise ArtifactMetadataError(message) from None
        else:
            data = parse_headers(metadata_file_contents)

    data['filetype'] = 'sdist'
    data['pyversion'] = 'source'

    return data


def parse_headers(metadata_file_contents):
    import email

    message = email.message_from_string(metadata_file_contents)

    headers = {'description': message.get_payload()}

    for header, value in message.items():
        normalized_header = header.lower().replace('-', '_')
        header_name = RENAMED_METADATA_FIELDS.get(normalized_header, normalized_header)

        if normalized_header in MULTIPLE_USE_METADATA_FIELDS:
            if header_name in headers:
                headers[header_name].append(value)
            else:
                headers[header_name] = [value]
        else:
            headers[header_name] = value

    return headers



# This file is dual licensed under the terms of the Apache License, Version
# 2.0, and the BSD License. See the LICENSE file in the root of this repository
# for complete details.

from collections.abc import Sequence

from cryptography.utils import Buffer

class AESGCM:
    def __init__(self, key: Buffer) -> None: ...
    @staticmethod
    def generate_key(bit_length: int) -> bytes: ...
    def encrypt(
        self,
        nonce: Buffer,
        data: Buffer,
        associated_data: Buffer | None,
    ) -> bytes: ...
    def decrypt(
        self,
        nonce: Buffer,
        data: Buffer,
        associated_data: Buffer | None,
    ) -> bytes: ...
    def encrypt_into(
        self,
        nonce: Buffer,
        data: Buffer,
        associated_data: Buffer | None,
        buf: Buffer,
    ) -> int: ...
    def decrypt_into(
        self,
        nonce: Buffer,
        data: Buffer,
        associated_data: Buffer | None,
        buf: Buffer,
    ) -> int: ...

class ChaCha20Poly1305:
    def __init__(self, key: Buffer) -> None: ...
    @staticmethod
    def generate_key() -> bytes: ...
    def encrypt(
        self,
        nonce: Buffer,
        data: Buffer,
        associated_data: Buffer | None,
    ) -> bytes: ...
    def encrypt_into(
        self,
        nonce: Buffer,
        data: Buffer,
        associated_data: Buffer | None,
        buf: Buffer,
    ) -> int: ...
    def decrypt(
        self,
        nonce: Buffer,
        data: Buffer,
        associated_data: Buffer | None,
    ) -> bytes: ...
    def decrypt_into(
        self,
        nonce: Buffer,
        data: Buffer,
        associated_data: Buffer | None,
        buf: Buffer,
    ) -> int: ...

class AESCCM:
    def __init__(self, key: Buffer, tag_length: int = 16) -> None: ...
    @staticmethod
    def generate_key(bit_length: int) -> bytes: ...
    def encrypt(
        self,
        nonce: Buffer,
        data: Buffer,
        associated_data: Buffer | None,
    ) -> bytes: ...
    def encrypt_into(
        self,
        nonce: Buffer,
        data: Buffer,
        associated_data: Buffer | None,
        buf: Buffer,
    ) -> int: ...
    def decrypt(
        self,
        nonce: Buffer,
        data: Buffer,
        associated_data: Buffer | None,
    ) -> bytes: ...
    def decrypt_into(
        self,
        nonce: Buffer,
        data: Buffer,
        associated_data: Buffer | None,
        buf: Buffer,
    ) -> int: ...

class AESSIV:
    def __init__(self, key: Buffer) -> None: ...
    @staticmethod
    def generate_key(bit_length: int) -> bytes: ...
    def encrypt(
        self,
        data: Buffer,
        associated_data: Sequence[Buffer] | None,
    ) -> bytes: ...
    def encrypt_into(
        self,
        data: Buffer,
        associated_data: Sequence[Buffer] | None,
        buf: Buffer,
    ) -> int: ...
    def decrypt(
        self,
        data: Buffer,
        associated_data: Sequence[Buffer] | None,
    ) -> bytes: ...
    def decrypt_into(
        self,
        data: Buffer,
        associated_data: Sequence[Buffer] | None,
        buf: Buffer,
    ) -> int: ...

class AESOCB3:
    def __init__(self, key: Buffer) -> None: ...
    @staticmethod
    def generate_key(bit_length: int) -> bytes: ...
    def encrypt(
        self,
        nonce: Buffer,
        data: Buffer,
        associated_data: Buffer | None,
    ) -> bytes: ...
    def encrypt_into(
        self,
        nonce: Buffer,
        data: Buffer,
        associated_data: Buffer | None,
        buf: Buffer,
    ) -> int: ...
    def decrypt(
        self,
        nonce: Buffer,
        data: Buffer,
        associated_data: Buffer | None,
    ) -> bytes: ...
    def decrypt_into(
        self,
        nonce: Buffer,
        data: Buffer,
        associated_data: Buffer | None,
        buf: Buffer,
    ) -> int: ...

class AESGCMSIV:
    def __init__(self, key: Buffer) -> None: ...
    @staticmethod
    def generate_key(bit_length: int) -> bytes: ...
    def encrypt(
        self,
        nonce: Buffer,
        data: Buffer,
        associated_data: Buffer | None,
    ) -> bytes: ...
    def encrypt_into(
        self,
        nonce: Buffer,
        data: Buffer,
        associated_data: Buffer | None,
        buf: Buffer,
    ) -> int: ...
    def decrypt(
        self,
        nonce: Buffer,
        data: Buffer,
        associated_data: Buffer | None,
    ) -> bytes: ...
    def decrypt_into(
        self,
        nonce: Buffer,
        data: Buffer,
        associated_data: Buffer | None,
        buf: Buffer,
    ) -> int: ...



import click


@click.command(short_help='Create or initialize a project')
@click.argument('name', required=False)
@click.argument('location', required=False)
@click.option('--interactive', '-i', 'interactive', is_flag=True, help='Interactively choose details about the project')
@click.option('--cli', 'feature_cli', is_flag=True, help='Give the project a command line interface')
@click.option('--init', 'initialize', is_flag=True, help='Initialize an existing project')
@click.option('-so', 'setuptools_options', multiple=True, hidden=True)
@click.pass_obj
def new(app, name, location, interactive, feature_cli, initialize, setuptools_options):
    """Create or initialize a project."""
    import sys
    from copy import deepcopy
    from datetime import datetime, timezone

    from hatch.config.model import TemplateConfig
    from hatch.project.core import Project
    from hatch.template import File
    from hatch.utils.fs import Path

    if location:
        location = Path(location).resolve()

    migration_possible = False
    if initialize:
        interactive = True
        location = location or Path.cwd()
        if (location / 'setup.py').is_file() or (location / 'setup.cfg').is_file():
            migration_possible = True
            if not name:
                name = 'temporary'

    if not name:
        if not interactive:
            app.abort('Missing required argument for the project name, use the -i/--interactive flag.')

        name = app.prompt('Project name')

    normalized_name = Project.canonicalize_name(name, strict=False)
    if not location:
        location = Path(normalized_name).resolve()

    needs_config_update = False
    if location.is_file():
        app.abort(f'Path `{location}` points to a file.')
    elif location.is_dir() and any(location.iterdir()):
        if not initialize:
            app.abort(f'Directory `{location}` is not empty.')

        needs_config_update = (location / 'pyproject.toml').is_file()

    if migration_possible:
        from hatch.cli.new.migrate import migrate
        from hatch.venv.core import TempVirtualEnv

        try:
            with app.status('Migrating project metadata from setuptools'), TempVirtualEnv(
                sys.executable, app.platform
            ) as venv:
                app.platform.run_command(['python', '-m', 'pip', 'install', '-q', 'setuptools'])
                migrate(str(location), setuptools_options, venv.sys_path)
        except Exception as e:  # noqa: BLE001
            app.display_error(f'Could not automatically migrate from setuptools: {e}')
            if name == 'temporary':
                name = app.prompt('Project name')
        else:
            return

    default_config = {
        'description': '',
        'dependencies': set(),
        'package_name': normalized_name.replace('-', '_'),
        'project_name': name,
        'project_name_normalized': normalized_name,
        'args': {'cli': feature_cli},
    }

    if interactive:
        default_config['description'] = app.prompt('Description', default='')

        app.display_info()

    if needs_config_update:
        app.project.initialize(str(location / 'pyproject.toml'), default_config)
        app.display_success('Updated: pyproject.toml')
        return

    template_config = deepcopy(app.config.template.raw_data)
    if 'plugins' in template_config and not template_config['plugins']:
        del template_config['plugins']

    TemplateConfig(template_config, ('template',)).parse_fields()

    plugin_config = template_config.pop('plugins')

    # Set up default config for template files
    template_config.update(default_config)

    template_classes = app.plugins.template.collect()

    templates = []
    for template_name, template_class in sorted(template_classes.items(), key=lambda item: -item[1].PRIORITY):
        if template_name in plugin_config:
            templates.append(
                template_class(plugin_config.pop(template_name), app.cache_dir, datetime.now(timezone.utc))
            )

    if not templates:
        app.abort(f'None of the defined plugins were found: {", ".join(sorted(plugin_config))}')
    elif plugin_config:
        app.abort(f'Some of the defined plugins were not found: {", ".join(sorted(plugin_config))}')

    for template in templates:
        template.initialize_config(template_config)

    template_files = []

    for template in templates:
        for possible_template_file in template.get_files(config=deepcopy(template_config)):
            template_file = (
                possible_template_file(deepcopy(template_config), template.plugin_config)
                if possible_template_file.__class__ is not File
                else possible_template_file
            )
            if template_file.path is None or (initialize and str(template_file.path) != 'pyproject.toml'):
                continue

            template_files.append(template_file)

    for template in templates:
        template.finalize_files(config=deepcopy(template_config), files=template_files)

    for template_file in template_files:
        template_file.write(location)

    if initialize:
        app.display_success('Wrote: pyproject.toml')
        return

    from rich.markup import escape
    from rich.tree import Tree

    def recurse_directory(directory, tree):
        paths = sorted(Path(directory).iterdir(), key=lambda p: (p.is_file(), p.name))
        for path in paths:
            if path.is_dir():
                recurse_directory(
                    path, tree.add(f'[bold magenta][link={app.platform.format_file_uri(path)}]{escape(path.name)}')
                )
            else:
                tree.add(f'[green][link={app.platform.format_file_uri(path)}]{escape(path.name)}')

    root = Tree(
        f'[bold magenta][link={app.platform.format_file_uri(location)}]{escape(location.name)}',
        guide_style='bright_blue',
        hide_root=location == Path.cwd(),
    )
    recurse_directory(location, root)
    app.output(root, markup=True)



import os

FIELD_TO_PARSE = object()


class ConfigurationError(Exception):
    def __init__(self, *args, location):
        self.location = location
        super().__init__(*args)

    def __str__(self):
        return f'Error parsing config:\n{self.location}\n  {super().__str__()}'


def parse_config(obj):
    if isinstance(obj, LazilyParsedConfig):
        obj.parse_fields()
    elif isinstance(obj, list):
        for o in obj:
            parse_config(o)
    elif isinstance(obj, dict):
        for o in obj.values():
            parse_config(o)


class LazilyParsedConfig:
    def __init__(self, config: dict, steps: tuple = ()):
        self.raw_data = config
        self.steps = steps

    def parse_fields(self):
        for attribute in self.__dict__:
            _, prefix, name = attribute.partition('_field_')
            if prefix:
                parse_config(getattr(self, name))

    def raise_error(self, message, *, extra_steps=()):
        import inspect

        field = inspect.currentframe().f_back.f_code.co_name
        raise ConfigurationError(message, location=' -> '.join([*self.steps, field, *extra_steps]))


class RootConfig(LazilyParsedConfig):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._field_mode = FIELD_TO_PARSE
        self._field_project = FIELD_TO_PARSE
        self._field_shell = FIELD_TO_PARSE
        self._field_dirs = FIELD_TO_PARSE
        self._field_projects = FIELD_TO_PARSE
        self._field_publish = FIELD_TO_PARSE
        self._field_template = FIELD_TO_PARSE
        self._field_terminal = FIELD_TO_PARSE

    @property
    def mode(self):
        if self._field_mode is FIELD_TO_PARSE:
            if 'mode' in self.raw_data:
                mode = self.raw_data['mode']
                if not isinstance(mode, str):
                    self.raise_error('must be a string')

                valid_modes = ('aware', 'local', 'project')
                if mode not in valid_modes:
                    self.raise_error(f'must be one of: {", ".join(valid_modes)}')

                self._field_mode = mode
            else:
                self._field_mode = self.raw_data['mode'] = 'local'

        return self._field_mode

    @mode.setter
    def mode(self, value):
        self.raw_data['mode'] = value
        self._field_mode = FIELD_TO_PARSE

    @property
    def project(self):
        if self._field_project is FIELD_TO_PARSE:
            if 'project' in self.raw_data:
                project = self.raw_data['project']
                if not isinstance(project, str):
                    self.raise_error('must be a string')

                self._field_project = project
            else:
                self._field_project = self.raw_data['project'] = ''

        return self._field_project

    @project.setter
    def project(self, value):
        self.raw_data['project'] = value
        self._field_project = FIELD_TO_PARSE

    @property
    def shell(self):
        if self._field_shell is FIELD_TO_PARSE:
            if 'shell' in self.raw_data:
                shell = self.raw_data['shell']
                if isinstance(shell, str):
                    self._field_shell = ShellConfig({'name': shell}, ('shell',))
                elif isinstance(shell, dict):
                    self._field_shell = ShellConfig(shell, ('shell',))
                else:
                    self.raise_error('must be a string or table')
            else:
                self.raw_data['shell'] = ''
                self._field_shell = ShellConfig({'name': ''}, ('shell',))

        return self._field_shell

    @shell.setter
    def shell(self, value):
        self.raw_data['shell'] = value
        self._field_shell = FIELD_TO_PARSE

    @property
    def dirs(self):
        if self._field_dirs is FIELD_TO_PARSE:
            if 'dirs' in self.raw_data:
                dirs = self.raw_data['dirs']
                if not isinstance(dirs, dict):
                    self.raise_error('must be a table')

                self._field_dirs = DirsConfig(dirs, ('dirs',))
            else:
                dirs = {}
                self.raw_data['dirs'] = dirs
                self._field_dirs = DirsConfig(dirs, ('dirs',))

        return self._field_dirs

    @dirs.setter
    def dirs(self, value):
        self.raw_data['dirs'] = value
        self._field_dirs = FIELD_TO_PARSE

    @property
    def projects(self):
        if self._field_projects is FIELD_TO_PARSE:
            if 'projects' in self.raw_data:
                projects = self.raw_data['projects']
                if not isinstance(projects, dict):
                    self.raise_error('must be a table')

                project_data = {}
                for name, data in projects.items():
                    if isinstance(data, str):
                        project_data[name] = ProjectConfig({'location': data}, ('projects', name))
                    elif isinstance(data, dict):
                        project_data[name] = ProjectConfig(data, ('projects', name))
                    else:
                        self.raise_error('must be a string or table', extra_steps=(name,))

                self._field_projects = project_data
            else:
                self._field_projects = self.raw_data['projects'] = {}

        return self._field_projects

    @projects.setter
    def projects(self, value):
        self.raw_data['projects'] = value
        self._field_projects = FIELD_TO_PARSE

    @property
    def publish(self):
        if self._field_publish is FIELD_TO_PARSE:
            if 'publish' in self.raw_data:
                publish = self.raw_data['publish']
                if not isinstance(publish, dict):
                    self.raise_error('must be a table')

                for name, data in publish.items():
                    if not isinstance(data, dict):
                        self.raise_error('must be a table', extra_steps=(name,))

                self._field_publish = publish
            else:
                self._field_publish = self.raw_data['publish'] = {'index': {'repo': 'main'}}

        return self._field_publish

    @publish.setter
    def publish(self, value):
        self.raw_data['publish'] = value
        self._field_publish = FIELD_TO_PARSE

    @property
    def template(self):
        if self._field_template is FIELD_TO_PARSE:
            if 'template' in self.raw_data:
                template = self.raw_data['template']
                if not isinstance(template, dict):
                    self.raise_error('must be a table')

                self._field_template = TemplateConfig(template, ('template',))
            else:
                template = {}
                self.raw_data['template'] = template
                self._field_template = TemplateConfig(template, ('template',))

        return self._field_template

    @template.setter
    def template(self, value):
        self.raw_data['template'] = value
        self._field_template = FIELD_TO_PARSE

    @property
    def terminal(self):
        if self._field_terminal is FIELD_TO_PARSE:
            if 'terminal' in self.raw_data:
                terminal = self.raw_data['terminal']
                if not isinstance(terminal, dict):
                    self.raise_error('must be a table')

                self._field_terminal = TerminalConfig(terminal, ('terminal',))
            else:
                terminal = {}
                self.raw_data['terminal'] = terminal
                self._field_terminal = TerminalConfig(terminal, ('terminal',))

        return self._field_terminal

    @terminal.setter
    def terminal(self, value):
        self.raw_data['terminal'] = value
        self._field_terminal = FIELD_TO_PARSE


class ShellConfig(LazilyParsedConfig):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._field_name = FIELD_TO_PARSE
        self._field_path = FIELD_TO_PARSE
        self._field_args = FIELD_TO_PARSE

    @property
    def name(self):
        if self._field_name is FIELD_TO_PARSE:
            if 'name' in self.raw_data:
                name = self.raw_data['name']
                if not isinstance(name, str):
                    self.raise_error('must be a string')

                self._field_name = name
            else:
                self.raise_error('required field')

        return self._field_name

    @name.setter
    def name(self, value):
        self.raw_data['name'] = value
        self._field_name = FIELD_TO_PARSE

    @property
    def path(self):
        if self._field_path is FIELD_TO_PARSE:
            if 'path' in self.raw_data:
                path = self.raw_data['path']
                if not isinstance(path, str):
                    self.raise_error('must be a string')

                self._field_path = path
            else:
                self._field_path = self.raw_data['path'] = self.name

        return self._field_path

    @path.setter
    def path(self, value):
        self.raw_data['path'] = value
        self._field_path = FIELD_TO_PARSE

    @property
    def args(self):
        if self._field_args is FIELD_TO_PARSE:
            if 'args' in self.raw_data:
                args = self.raw_data['args']
                if not isinstance(args, list):
                    self.raise_error('must be an array')

                for i, entry in enumerate(args, 1):
                    if not isinstance(entry, str):
                        self.raise_error('must be a string', extra_steps=(str(i),))

                self._field_args = args
            else:
                self._field_args = self.raw_data['args'] = []

        return self._field_args

    @args.setter
    def args(self, value):
        self.raw_data['args'] = value
        self._field_args = FIELD_TO_PARSE


class DirsConfig(LazilyParsedConfig):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._field_project = FIELD_TO_PARSE
        self._field_env = FIELD_TO_PARSE
        self._field_python = FIELD_TO_PARSE
        self._field_data = FIELD_TO_PARSE
        self._field_cache = FIELD_TO_PARSE

    @property
    def project(self):
        if self._field_project is FIELD_TO_PARSE:
            if 'project' in self.raw_data:
                project = self.raw_data['project']
                if not isinstance(project, list):
                    self.raise_error('must be an array')

                for i, entry in enumerate(project, 1):
                    if not isinstance(entry, str):
                        self.raise_error('must be a string', extra_steps=(str(i),))

                self._field_project = project
            else:
                self._field_project = self.raw_data['project'] = []

        return self._field_project

    @project.setter
    def project(self, value):
        self.raw_data['project'] = value
        self._field_project = FIELD_TO_PARSE

    @property
    def env(self):
        if self._field_env is FIELD_TO_PARSE:
            if 'env' in self.raw_data:
                env = self.raw_data['env']
                if not isinstance(env, dict):
                    self.raise_error('must be a table')

                for key, value in env.items():
                    if not isinstance(value, str):
                        self.raise_error('must be a string', extra_steps=(key,))

                self._field_env = env
            else:
                self._field_env = self.raw_data['env'] = {}

        return self._field_env

    @env.setter
    def env(self, value):
        self.raw_data['env'] = value
        self._field_env = FIELD_TO_PARSE

    @property
    def python(self):
        if self._field_python is FIELD_TO_PARSE:
            if 'python' in self.raw_data:
                python = self.raw_data['python']
                if not isinstance(python, str):
                    self.raise_error('must be a string')

                self._field_python = python
            else:
                self._field_python = self.raw_data['python'] = 'isolated'

        return self._field_python

    @python.setter
    def python(self, value):
        self.raw_data['python'] = value
        self._field_python = FIELD_TO_PARSE

    @property
    def data(self):
        if self._field_data is FIELD_TO_PARSE:
            if 'data' in self.raw_data:
                data = self.raw_data['data']
                if not isinstance(data, str):
                    self.raise_error('must be a string')

                self._field_data = data
            else:
                from platformdirs import user_data_dir

                self._field_data = self.raw_data['data'] = user_data_dir('hatch', appauthor=False)

        return self._field_data

    @data.setter
    def data(self, value):
        self.raw_data['data'] = value
        self._field_data = FIELD_TO_PARSE

    @property
    def cache(self):
        if self._field_cache is FIELD_TO_PARSE:
            if 'cache' in self.raw_data:
                cache = self.raw_data['cache']
                if not isinstance(cache, str):
                    self.raise_error('must be a string')

                self._field_cache = cache
            else:
                from platformdirs import user_cache_dir

                self._field_cache = self.raw_data['cache'] = user_cache_dir('hatch', appauthor=False)

        return self._field_cache

    @cache.setter
    def cache(self, value):
        self.raw_data['cache'] = value
        self._field_cache = FIELD_TO_PARSE


class ProjectConfig(LazilyParsedConfig):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._field_location = FIELD_TO_PARSE

    @property
    def location(self):
        if self._field_location is FIELD_TO_PARSE:
            if 'location' in self.raw_data:
                location = self.raw_data['location']
                if not isinstance(location, str):
                    self.raise_error('must be a string')

                self._field_location = location
            else:
                self.raise_error('required field')

        return self._field_location

    @location.setter
    def location(self, value):
        self.raw_data['location'] = value
        self._field_location = FIELD_TO_PARSE


class TemplateConfig(LazilyParsedConfig):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._field_name = FIELD_TO_PARSE
        self._field_email = FIELD_TO_PARSE
        self._field_licenses = FIELD_TO_PARSE
        self._field_plugins = FIELD_TO_PARSE

    @property
    def name(self):
        if self._field_name is FIELD_TO_PARSE:
            if 'name' in self.raw_data:
                name = self.raw_data['name']
                if not isinstance(name, str):
                    self.raise_error('must be a string')

                self._field_name = name
            else:
                name = os.environ.get('GIT_AUTHOR_NAME')
                if name is None:
                    import subprocess

                    try:
                        name = subprocess.check_output(
                            ['git', 'config', '--get', 'user.name'],  # noqa: S607
                            text=True,
                        ).strip()
                    except Exception:  # noqa: BLE001
                        name = 'U.N. Owen'

                self._field_name = self.raw_data['name'] = name

        return self._field_name

    @name.setter
    def name(self, value):
        self.raw_data['name'] = value
        self._field_name = FIELD_TO_PARSE

    @property
    def email(self):
        if self._field_email is FIELD_TO_PARSE:
            if 'email' in self.raw_data:
                email = self.raw_data['email']
                if not isinstance(email, str):
                    self.raise_error('must be a string')

                self._field_email = email
            else:
                email = os.environ.get('GIT_AUTHOR_EMAIL')
                if email is None:
                    import subprocess

                    try:
                        email = subprocess.check_output(
                            ['git', 'config', '--get', 'user.email'],  # noqa: S607
                            text=True,
                        ).strip()
                    except Exception:  # noqa: BLE001
                        email = 'void@some.where'

                self._field_email = self.raw_data['email'] = email

        return self._field_email

    @email.setter
    def email(self, value):
        self.raw_data['email'] = value
        self._field_email = FIELD_TO_PARSE

    @property
    def licenses(self):
        if self._field_licenses is FIELD_TO_PARSE:
            if 'licenses' in self.raw_data:
                licenses = self.raw_data['licenses']
                if not isinstance(licenses, dict):
                    self.raise_error('must be a table')

                self._field_licenses = LicensesConfig(licenses, (*self.steps, 'licenses'))
            else:
                licenses = {}
                self.raw_data['licenses'] = licenses
                self._field_licenses = LicensesConfig(licenses, (*self.steps, 'licenses'))

        return self._field_licenses

    @licenses.setter
    def licenses(self, value):
        self.raw_data['licenses'] = value
        self._field_licenses = FIELD_TO_PARSE

    @property
    def plugins(self):
        if self._field_plugins is FIELD_TO_PARSE:
            if 'plugins' in self.raw_data:
                plugins = self.raw_data['plugins']
                if not isinstance(plugins, dict):
                    self.raise_error('must be a table')

                for name, data in plugins.items():
                    if not isinstance(data, dict):
                        self.raise_error('must be a table', extra_steps=(name,))

                self._field_plugins = plugins
            else:
                self._field_plugins = self.raw_data['plugins'] = {
                    'default': {'tests': True, 'ci': False, 'src-layout': True}
                }

        return self._field_plugins

    @plugins.setter
    def plugins(self, value):
        self.raw_data['plugins'] = value
        self._field_plugins = FIELD_TO_PARSE


class LicensesConfig(LazilyParsedConfig):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._field_headers = FIELD_TO_PARSE
        self._field_default = FIELD_TO_PARSE

    @property
    def headers(self):
        if self._field_headers is FIELD_TO_PARSE:
            if 'headers' in self.raw_data:
                headers = self.raw_data['headers']
                if not isinstance(headers, bool):
                    self.raise_error('must be a boolean')

                self._field_headers = headers
            else:
                self._field_headers = self.raw_data['headers'] = True

        return self._field_headers

    @headers.setter
    def headers(self, value):
        self.raw_data['headers'] = value
        self._field_headers = FIELD_TO_PARSE

    @property
    def default(self):
        if self._field_default is FIELD_TO_PARSE:
            if 'default' in self.raw_data:
                default = self.raw_data['default']
                if not isinstance(default, list):
                    self.raise_error('must be an array')

                for i, entry in enumerate(default, 1):
                    if not isinstance(entry, str):
                        self.raise_error('must be a string', extra_steps=(str(i),))

                self._field_default = default
            else:
                self._field_default = self.raw_data['default'] = ['MIT']

        return self._field_default

    @default.setter
    def default(self, value):
        self.raw_data['default'] = value
        self._field_default = FIELD_TO_PARSE


class TerminalConfig(LazilyParsedConfig):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._field_styles = FIELD_TO_PARSE

    @property
    def styles(self):
        if self._field_styles is FIELD_TO_PARSE:
            if 'styles' in self.raw_data:
                styles = self.raw_data['styles']
                if not isinstance(styles, dict):
                    self.raise_error('must be a table')

                self._field_styles = StylesConfig(styles, (*self.steps, 'styles'))
            else:
                styles = {}
                self.raw_data['styles'] = styles
                self._field_styles = StylesConfig(styles, (*self.steps, 'styles'))

        return self._field_styles

    @styles.setter
    def styles(self, value):
        self.raw_data['styles'] = value
        self._field_styles = FIELD_TO_PARSE


class StylesConfig(LazilyParsedConfig):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._field_info = FIELD_TO_PARSE
        self._field_success = FIELD_TO_PARSE
        self._field_error = FIELD_TO_PARSE
        self._field_warning = FIELD_TO_PARSE
        self._field_waiting = FIELD_TO_PARSE
        self._field_debug = FIELD_TO_PARSE
        self._field_spinner = FIELD_TO_PARSE

    @property
    def info(self):
        if self._field_info is FIELD_TO_PARSE:
            if 'info' in self.raw_data:
                info = self.raw_data['info']
                if not isinstance(info, str):
                    self.raise_error('must be a string')

                self._field_info = info
            else:
                self._field_info = self.raw_data['info'] = 'bold'

        return self._field_info

    @info.setter
    def info(self, value):
        self.raw_data['info'] = value
        self._field_info = FIELD_TO_PARSE

    @property
    def success(self):
        if self._field_success is FIELD_TO_PARSE:
            if 'success' in self.raw_data:
                success = self.raw_data['success']
                if not isinstance(success, str):
                    self.raise_error('must be a string')

                self._field_success = success
            else:
                self._field_success = self.raw_data['success'] = 'bold cyan'

        return self._field_success

    @success.setter
    def success(self, value):
        self.raw_data['success'] = value
        self._field_success = FIELD_TO_PARSE

    @property
    def error(self):
        if self._field_error is FIELD_TO_PARSE:
            if 'error' in self.raw_data:
                error = self.raw_data['error']
                if not isinstance(error, str):
                    self.raise_error('must be a string')

                self._field_error = error
            else:
                self._field_error = self.raw_data['error'] = 'bold red'

        return self._field_error

    @error.setter
    def error(self, value):
        self.raw_data['error'] = value
        self._field_error = FIELD_TO_PARSE

    @property
    def warning(self):
        if self._field_warning is FIELD_TO_PARSE:
            if 'warning' in self.raw_data:
                warning = self.raw_data['warning']
                if not isinstance(warning, str):
                    self.raise_error('must be a string')

                self._field_warning = warning
            else:
                self._field_warning = self.raw_data['warning'] = 'bold yellow'

        return self._field_warning

    @warning.setter
    def warning(self, value):
        self.raw_data['warning'] = value
        self._field_warning = FIELD_TO_PARSE

    @property
    def waiting(self):
        if self._field_waiting is FIELD_TO_PARSE:
            if 'waiting' in self.raw_data:
                waiting = self.raw_data['waiting']
                if not isinstance(waiting, str):
                    self.raise_error('must be a string')

                self._field_waiting = waiting
            else:
                self._field_waiting = self.raw_data['waiting'] = 'bold magenta'

        return self._field_waiting

    @waiting.setter
    def waiting(self, value):
        self.raw_data['waiting'] = value
        self._field_waiting = FIELD_TO_PARSE

    @property
    def debug(self):
        if self._field_debug is FIELD_TO_PARSE:
            if 'debug' in self.raw_data:
                debug = self.raw_data['debug']
                if not isinstance(debug, str):
                    self.raise_error('must be a string')

                self._field_debug = debug
            else:
                self._field_debug = self.raw_data['debug'] = 'bold'

        return self._field_debug

    @debug.setter
    def debug(self, value):
        self.raw_data['debug'] = value
        self._field_debug = FIELD_TO_PARSE

    @property
    def spinner(self):
        if self._field_spinner is FIELD_TO_PARSE:
            if 'spinner' in self.raw_data:
                spinner = self.raw_data['spinner']
                if not isinstance(spinner, str):
                    self.raise_error('must be a string')

                self._field_spinner = spinner
            else:
                self._field_spinner = self.raw_data['spinner'] = 'simpleDotsScrolling'

        return self._field_spinner

    @spinner.setter
    def spinner(self, value):
        self.raw_data['spinner'] = value
        self._field_spinner = FIELD_TO_PARSE



"""Tests for distutils.command.install_data."""

import importlib.util
import os
import sys
from distutils.command.install_lib import install_lib
from distutils.errors import DistutilsOptionError
from distutils.extension import Extension
from distutils.tests import support

import pytest


@support.combine_markers
@pytest.mark.usefixtures('save_env')
class TestInstallLib(
    support.TempdirManager,
):
    def test_finalize_options(self):
        dist = self.create_dist()[1]
        cmd = install_lib(dist)

        cmd.finalize_options()
        assert cmd.compile == 1
        assert cmd.optimize == 0

        # optimize must be 0, 1, or 2
        cmd.optimize = 'foo'
        with pytest.raises(DistutilsOptionError):
            cmd.finalize_options()
        cmd.optimize = '4'
        with pytest.raises(DistutilsOptionError):
            cmd.finalize_options()

        cmd.optimize = '2'
        cmd.finalize_options()
        assert cmd.optimize == 2

    @pytest.mark.skipif('sys.dont_write_bytecode')
    def test_byte_compile(self):
        project_dir, dist = self.create_dist()
        os.chdir(project_dir)
        cmd = install_lib(dist)
        cmd.compile = cmd.optimize = 1

        f = os.path.join(project_dir, 'foo.py')
        self.write_file(f, '# python file')
        cmd.byte_compile([f])
        pyc_file = importlib.util.cache_from_source('foo.py', optimization='')
        pyc_opt_file = importlib.util.cache_from_source(
            'foo.py', optimization=cmd.optimize
        )
        assert os.path.exists(pyc_file)
        assert os.path.exists(pyc_opt_file)

    def test_get_outputs(self):
        project_dir, dist = self.create_dist()
        os.chdir(project_dir)
        os.mkdir('spam')
        cmd = install_lib(dist)

        # setting up a dist environment
        cmd.compile = cmd.optimize = 1
        cmd.install_dir = self.mkdtemp()
        f = os.path.join(project_dir, 'spam', '__init__.py')
        self.write_file(f, '# python package')
        cmd.distribution.ext_modules = [Extension('foo', ['xxx'])]
        cmd.distribution.packages = ['spam']
        cmd.distribution.script_name = 'setup.py'

        # get_outputs should return 4 elements: spam/__init__.py and .pyc,
        # foo.import-tag-abiflags.so / foo.pyd
        outputs = cmd.get_outputs()
        assert len(outputs) == 4, outputs

    def test_get_inputs(self):
        project_dir, dist = self.create_dist()
        os.chdir(project_dir)
        os.mkdir('spam')
        cmd = install_lib(dist)

        # setting up a dist environment
        cmd.compile = cmd.optimize = 1
        cmd.install_dir = self.mkdtemp()
        f = os.path.join(project_dir, 'spam', '__init__.py')
        self.write_file(f, '# python package')
        cmd.distribution.ext_modules = [Extension('foo', ['xxx'])]
        cmd.distribution.packages = ['spam']
        cmd.distribution.script_name = 'setup.py'

        # get_inputs should return 2 elements: spam/__init__.py and
        # foo.import-tag-abiflags.so / foo.pyd
        inputs = cmd.get_inputs()
        assert len(inputs) == 2, inputs

    def test_dont_write_bytecode(self, caplog):
        # makes sure byte_compile is not used
        dist = self.create_dist()[1]
        cmd = install_lib(dist)
        cmd.compile = True
        cmd.optimize = 1

        old_dont_write_bytecode = sys.dont_write_bytecode
        sys.dont_write_bytecode = True
        try:
            cmd.byte_compile([])
        finally:
            sys.dont_write_bytecode = old_dont_write_bytecode

        assert 'byte-compiling is disabled' in caplog.messages[0]



from __future__ import annotations

import re
from typing import Iterable

from hatch.publish.plugin.interface import PublisherInterface
from hatch.utils.fs import Path
from hatchling.metadata.utils import normalize_project_name


class IndexPublisher(PublisherInterface):
    PLUGIN_NAME = 'index'

    def get_repos(self):
        global_plugin_config = self.plugin_config.copy()
        defined_repos = self.plugin_config.pop('repos', {})
        self.plugin_config.pop('repo', None)

        # Normalize type
        repos = {}
        for repo, data in defined_repos.items():
            if isinstance(data, str):
                repos[repo] = {'url': data}
            elif not isinstance(data, dict):
                self.app.abort(f'Hatch config field `publish.index.repos.{repo}` must be a string or a mapping')
            elif 'url' not in data:
                self.app.abort(f'Hatch config field `publish.index.repos.{repo}` must define a `url` key')
            else:
                repos[repo] = data

        # Ensure PyPI correct
        for repo, url in (
            ('main', 'https://upload.pypi.org/legacy/'),
            ('test', 'https://test.pypi.org/legacy/'),
        ):
            repos.setdefault(repo, {})['url'] = url

        # Populate defaults
        for config in repos.values():
            for key, value in global_plugin_config.items():
                config.setdefault(key, value)

        return repos

    def publish(self, artifacts: list, options: dict):
        """
        https://warehouse.readthedocs.io/api-reference/legacy.html#upload-api
        """
        from collections import defaultdict

        from hatch.index.core import PackageIndex
        from hatch.index.publish import get_sdist_form_data, get_wheel_form_data
        from hatch.publish.auth import AuthenticationCredentials

        if not artifacts:
            from hatchling.builders.constants import DEFAULT_BUILD_DIRECTORY

            artifacts = [DEFAULT_BUILD_DIRECTORY]

        repo = options['repo'] if 'repo' in options else self.plugin_config.get('repo', 'main')
        repos = self.get_repos()
        repo_config: dict[str, str] = repos[repo] if repo in repos else {'url': repo}
        credentials = AuthenticationCredentials(
            app=self.app,
            cache_dir=self.cache_dir,
            options=options,
            repo=repo,
            repo_config=repo_config,
        )

        index = PackageIndex(
            repo_config['url'],
            user=credentials.username,
            auth=credentials.password,
            ca_cert=options.get('ca_cert', repo_config.get('ca-cert')),
            client_cert=options.get('client_cert', repo_config.get('client-cert')),
            client_key=options.get('client_key', repo_config.get('client-key')),
        )

        existing_artifacts: dict[str, set[str]] = {}

        # Use as an ordered set
        project_versions: dict[str, dict[str, None]] = defaultdict(dict)

        artifacts_found = False
        for artifact in recurse_artifacts(artifacts, self.root):
            if artifact.name.endswith('.whl'):
                data = get_wheel_form_data(artifact)
            elif artifact.name.endswith('.tar.gz'):
                data = get_sdist_form_data(artifact)
            else:
                continue

            artifacts_found = True

            for field in ('name', 'version'):
                if field not in data:
                    self.app.abort(f'Missing required field `{field}` in artifact: {artifact}')

            try:
                displayed_path = str(artifact.relative_to(self.root))
            except ValueError:
                displayed_path = str(artifact)

            self.app.display_info(f'{displayed_path} ...', end=' ')

            project_name = normalize_project_name(data['name'])
            if project_name not in existing_artifacts:
                try:
                    response = index.get_simple_api(project_name)
                    response.raise_for_status()
                except Exception:  # no cov  # noqa: BLE001
                    existing_artifacts[project_name] = set()
                else:
                    existing_artifacts[project_name] = set(parse_artifacts(response.text))

            if artifact.name in existing_artifacts[project_name]:
                self.app.display_warning('already exists')
                continue

            try:
                index.upload_artifact(artifact, data)
            except Exception as e:  # noqa: BLE001
                self.app.display_error('failed')
                self.app.abort(f'Error uploading to repository: {index.repo} - {e}'.replace(index.auth, '*****'))
            else:
                self.app.display_success('success')

                existing_artifacts[project_name].add(artifact.name)
                project_versions[project_name][data['version']] = None

        if not options['initialize_auth']:
            if not artifacts_found:
                self.app.abort('No artifacts found')
            elif not project_versions:
                self.app.abort(code=0)

        for project_name, versions in project_versions.items():
            self.app.display_info()
            self.app.display_mini_header(project_name)
            for version in versions:
                self.app.display_info(str(index.urls.project.child(project_name, version, '').to_iri()))

        credentials.write_updated_data()


def recurse_artifacts(artifacts: list, root) -> Iterable[Path]:
    for raw_artifact in artifacts:
        artifact = Path(raw_artifact)
        if not artifact.is_absolute():
            artifact = root / artifact

        if artifact.is_file():
            yield artifact
        elif artifact.is_dir():
            yield from artifact.iterdir()


def parse_artifacts(artifact_payload):
    for match in re.finditer(r'<a [^>]+>([^<]+)</a>', artifact_payload):
        yield match.group(1)



"""Utilities for determining application-specific dirs.

Provides convenience functions (e.g. :func:`user_data_dir`, :func:`user_config_path`), a :data:`PlatformDirs` class that
auto-detects the current platform, and the :class:`~platformdirs.api.PlatformDirsABC` base class.

See <https://github.com/platformdirs/platformdirs> for details and usage.

"""

from __future__ import annotations

import os
import sys
from typing import TYPE_CHECKING

from .api import PlatformDirsABC
from .version import __version__
from .version import __version_tuple__ as __version_info__

if TYPE_CHECKING:
    from pathlib import Path
    from typing import Literal

if sys.platform == "win32":
    from platformdirs.windows import Windows as _Result
elif sys.platform == "darwin":
    from platformdirs.macos import MacOS as _Result
else:
    from platformdirs.unix import Unix as _Result


def _set_platform_dir_class() -> type[PlatformDirsABC]:
    if os.getenv("ANDROID_DATA") == "/data" and os.getenv("ANDROID_ROOT") == "/system":
        if os.getenv("SHELL") or os.getenv("PREFIX"):
            return _Result

        from platformdirs.android import _android_folder  # noqa: PLC0415

        if _android_folder() is not None:
            from platformdirs.android import Android  # noqa: PLC0415

            return Android  # return to avoid redefinition of a result

    return _Result


if TYPE_CHECKING:
    # Work around mypy issue: https://github.com/python/mypy/issues/10962
    PlatformDirs = _Result
else:
    PlatformDirs = _set_platform_dir_class()  #: Currently active platform
AppDirs = PlatformDirs  #: Backwards compatibility with appdirs


def user_data_dir(  # noqa: PLR0913, PLR0917
    appname: str | None = None,
    appauthor: str | Literal[False] | None = None,
    version: str | None = None,
    roaming: bool = False,  # noqa: FBT001, FBT002
    ensure_exists: bool = False,  # noqa: FBT001, FBT002
    use_site_for_root: bool = False,  # noqa: FBT001, FBT002
) -> str:
    """:param appname: See `appname <platformdirs.api.PlatformDirsABC.appname>`.
    :param appauthor: See `appauthor <platformdirs.api.PlatformDirsABC.appauthor>`.
    :param version: See `version <platformdirs.api.PlatformDirsABC.version>`.
    :param roaming: See `roaming <platformdirs.api.PlatformDirsABC.roaming>`.
    :param ensure_exists: See `ensure_exists <platformdirs.api.PlatformDirsABC.ensure_exists>`.
    :param use_site_for_root: See `use_site_for_root <platformdirs.api.PlatformDirsABC.use_site_for_root>`.

    :returns: data directory tied to the user

    """
    return PlatformDirs(
        appname=appname,
        appauthor=appauthor,
        version=version,
        roaming=roaming,
        ensure_exists=ensure_exists,
        use_site_for_root=use_site_for_root,
    ).user_data_dir


def site_data_dir(
    appname: str | None = None,
    appauthor: str | Literal[False] | None = None,
    version: str | None = None,
    multipath: bool = False,  # noqa: FBT001, FBT002
    ensure_exists: bool = False,  # noqa: FBT001, FBT002
) -> str:
    """:param appname: See `appname <platformdirs.api.PlatformDirsABC.appname>`.
    :param appauthor: See `appauthor <platformdirs.api.PlatformDirsABC.appauthor>`.
    :param version: See `version <platformdirs.api.PlatformDirsABC.version>`.
    :param multipath: See `multipath <platformdirs.api.PlatformDirsABC.multipath>`.
    :param ensure_exists: See `ensure_exists <platformdirs.api.PlatformDirsABC.ensure_exists>`.

    :returns: data directory shared by users

    """
    return PlatformDirs(
        appname=appname,
        appauthor=appauthor,
        version=version,
        multipath=multipath,
        ensure_exists=ensure_exists,
    ).site_data_dir


def user_config_dir(  # noqa: PLR0913, PLR0917
    appname: str | None = None,
    appauthor: str | Literal[False] | None = None,
    version: str | None = None,
    roaming: bool = False,  # noqa: FBT001, FBT002
    ensure_exists: bool = False,  # noqa: FBT001, FBT002
    use_site_for_root: bool = False,  # noqa: FBT001, FBT002
) -> str:
    """:param appname: See `appname <platformdirs.api.PlatformDirsABC.appname>`.
    :param appauthor: See `appauthor <platformdirs.api.PlatformDirsABC.appauthor>`.
    :param version: See `version <platformdirs.api.PlatformDirsABC.version>`.
    :param roaming: See `roaming <platformdirs.api.PlatformDirsABC.roaming>`.
    :param ensure_exists: See `ensure_exists <platformdirs.api.PlatformDirsABC.ensure_exists>`.
    :param use_site_for_root: See `use_site_for_root <platformdirs.api.PlatformDirsABC.use_site_for_root>`.

    :returns: config directory tied to the user

    """
    return PlatformDirs(
        appname=appname,
        appauthor=appauthor,
        version=version,
        roaming=roaming,
        ensure_exists=ensure_exists,
        use_site_for_root=use_site_for_root,
    ).user_config_dir


def site_config_dir(
    appname: str | None = None,
    appauthor: str | Literal[False] | None = None,
    version: str | None = None,
    multipath: bool = False,  # noqa: FBT001, FBT002
    ensure_exists: bool = False,  # noqa: FBT001, FBT002
) -> str:
    """:param appname: See `appname <platformdirs.api.PlatformDirsABC.appname>`.
    :param appauthor: See `appauthor <platformdirs.api.PlatformDirsABC.appauthor>`.
    :param version: See `version <platformdirs.api.PlatformDirsABC.version>`.
    :param multipath: See `multipath <platformdirs.api.PlatformDirsABC.multipath>`.
    :param ensure_exists: See `ensure_exists <platformdirs.api.PlatformDirsABC.ensure_exists>`.

    :returns: config directory shared by users

    """
    return PlatformDirs(
        appname=appname,
        appauthor=appauthor,
        version=version,
        multipath=multipath,
        ensure_exists=ensure_exists,
    ).site_config_dir


def user_cache_dir(  # noqa: PLR0913, PLR0917
    appname: str | None = None,
    appauthor: str | Literal[False] | None = None,
    version: str | None = None,
    opinion: bool = True,  # noqa: FBT001, FBT002
    ensure_exists: bool = False,  # noqa: FBT001, FBT002
    use_site_for_root: bool = False,  # noqa: FBT001, FBT002
) -> str:
    """:param appname: See `appname <platformdirs.api.PlatformDirsABC.appname>`.
    :param appauthor: See `appauthor <platformdirs.api.PlatformDirsABC.appauthor>`.
    :param version: See `version <platformdirs.api.PlatformDirsABC.version>`.
    :param opinion: See `opinion <platformdirs.api.PlatformDirsABC.opinion>`.
    :param ensure_exists: See `ensure_exists <platformdirs.api.PlatformDirsABC.ensure_exists>`.
    :param use_site_for_root: See `use_site_for_root <platformdirs.api.PlatformDirsABC.use_site_for_root>`.

    :returns: cache directory tied to the user

    """
    return PlatformDirs(
        appname=appname,
        appauthor=appauthor,
        version=version,
        opinion=opinion,
        ensure_exists=ensure_exists,
        use_site_for_root=use_site_for_root,
    ).user_cache_dir


def site_cache_dir(
    appname: str | None = None,
    appauthor: str | Literal[False] | None = None,
    version: str | None = None,
    opinion: bool = True,  # noqa: FBT001, FBT002
    ensure_exists: bool = False,  # noqa: FBT001, FBT002
) -> str:
    """:param appname: See `appname <platformdirs.api.PlatformDirsABC.appname>`.
    :param appauthor: See `appauthor <platformdirs.api.PlatformDirsABC.appauthor>`.
    :param version: See `version <platformdirs.api.PlatformDirsABC.version>`.
    :param opinion: See `opinion <platformdirs.api.PlatformDirsABC.opinion>`.
    :param ensure_exists: See `ensure_exists <platformdirs.api.PlatformDirsABC.ensure_exists>`.

    :returns: cache directory shared by users

    """
    return PlatformDirs(
        appname=appname,
        appauthor=appauthor,
        version=version,
        opinion=opinion,
        ensure_exists=ensure_exists,
    ).site_cache_dir


def user_state_dir(  # noqa: PLR0913, PLR0917
    appname: str | None = None,
    appauthor: str | Literal[False] | None = None,
    version: str | None = None,
    roaming: bool = False,  # noqa: FBT001, FBT002
    ensure_exists: bool = False,  # noqa: FBT001, FBT002
    use_site_for_root: bool = False,  # noqa: FBT001, FBT002
) -> str:
    """:param appname: See `appname <platformdirs.api.PlatformDirsABC.appname>`.
    :param appauthor: See `appauthor <platformdirs.api.PlatformDirsABC.appauthor>`.
    :param version: See `version <platformdirs.api.PlatformDirsABC.version>`.
    :param roaming: See `roaming <platformdirs.api.PlatformDirsABC.roaming>`.
    :param ensure_exists: See `ensure_exists <platformdirs.api.PlatformDirsABC.ensure_exists>`.
    :param use_site_for_root: See `use_site_for_root <platformdirs.api.PlatformDirsABC.use_site_for_root>`.

    :returns: state directory tied to the user

    """
    return PlatformDirs(
        appname=appname,
        appauthor=appauthor,
        version=version,
        roaming=roaming,
        ensure_exists=ensure_exists,
        use_site_for_root=use_site_for_root,
    ).user_state_dir


def site_state_dir(
    appname: str | None = None,
    appauthor: str | Literal[False] | None = None,
    version: str | None = None,
    ensure_exists: bool = False,  # noqa: FBT001, FBT002
) -> str:
    """:param appname: See `appname <platformdirs.api.PlatformDirsABC.appname>`.
    :param appauthor: See `appauthor <platformdirs.api.PlatformDirsABC.appauthor>`.
    :param version: See `version <platformdirs.api.PlatformDirsABC.version>`.
    :param ensure_exists: See `ensure_exists <platformdirs.api.PlatformDirsABC.ensure_exists>`.

    :returns: state directory shared by users

    """
    return PlatformDirs(
        appname=appname,
        appauthor=appauthor,
        version=version,
        ensure_exists=ensure_exists,
    ).site_state_dir


def user_log_dir(  # noqa: PLR0913, PLR0917
    appname: str | None = None,
    appauthor: str | Literal[False] | None = None,
    version: str | None = None,
    opinion: bool = True,  # noqa: FBT001, FBT002
    ensure_exists: bool = False,  # noqa: FBT001, FBT002
    use_site_for_root: bool = False,  # noqa: FBT001, FBT002
) -> str:
    """:param appname: See `appname <platformdirs.api.PlatformDirsABC.appname>`.
    :param appauthor: See `appauthor <platformdirs.api.PlatformDirsABC.appauthor>`.
    :param version: See `version <platformdirs.api.PlatformDirsABC.version>`.
    :param opinion: See `opinion <platformdirs.api.PlatformDirsABC.opinion>`.
    :param ensure_exists: See `ensure_exists <platformdirs.api.PlatformDirsABC.ensure_exists>`.
    :param use_site_for_root: See `use_site_for_root <platformdirs.api.PlatformDirsABC.use_site_for_root>`.

    :returns: log directory tied to the user

    """
    return PlatformDirs(
        appname=appname,
        appauthor=appauthor,
        version=version,
        opinion=opinion,
        ensure_exists=ensure_exists,
        use_site_for_root=use_site_for_root,
    ).user_log_dir


def site_log_dir(
    appname: str | None = None,
    appauthor: str | Literal[False] | None = None,
    version: str | None = None,
    opinion: bool = True,  # noqa: FBT001, FBT002
    ensure_exists: bool = False,  # noqa: FBT001, FBT002
) -> str:
    """:param appname: See `appname <platformdirs.api.PlatformDirsABC.appname>`.
    :param appauthor: See `appauthor <platformdirs.api.PlatformDirsABC.appauthor>`.
    :param version: See `version <platformdirs.api.PlatformDirsABC.version>`.
    :param opinion: See `opinion <platformdirs.api.PlatformDirsABC.opinion>`.
    :param ensure_exists: See `ensure_exists <platformdirs.api.PlatformDirsABC.ensure_exists>`.

    :returns: log directory shared by users

    """
    return PlatformDirs(
        appname=appname,
        appauthor=appauthor,
        version=version,
        opinion=opinion,
        ensure_exists=ensure_exists,
    ).site_log_dir


def user_documents_dir() -> str:
    """:returns: documents directory tied to the user"""
    return PlatformDirs().user_documents_dir


def user_downloads_dir() -> str:
    """:returns: downloads directory tied to the user"""
    return PlatformDirs().user_downloads_dir


def user_pictures_dir() -> str:
    """:returns: pictures directory tied to the user"""
    return PlatformDirs().user_pictures_dir


def user_videos_dir() -> str:
    """:returns: videos directory tied to the user"""
    return PlatformDirs().user_videos_dir


def user_music_dir() -> str:
    """:returns: music directory tied to the user"""
    return PlatformDirs().user_music_dir


def user_desktop_dir() -> str:
    """:returns: desktop directory tied to the user"""
    return PlatformDirs().user_desktop_dir


def user_projects_dir() -> str:
    """:returns: projects directory tied to the user"""
    return PlatformDirs().user_projects_dir


def user_publicshare_dir() -> str:
    """:returns: public share directory tied to the user"""
    return PlatformDirs().user_publicshare_dir


def user_templates_dir() -> str:
    """:returns: templates directory tied to the user"""
    return PlatformDirs().user_templates_dir


def user_fonts_dir() -> str:
    """:returns: fonts directory tied to the user"""
    return PlatformDirs().user_fonts_dir


def user_preference_dir() -> str:
    """:returns: preference directory tied to the user"""
    return PlatformDirs().user_preference_dir


def user_bin_dir() -> str:
    """:returns: bin directory tied to the user"""
    return PlatformDirs().user_bin_dir


def site_bin_dir() -> str:
    """:returns: bin directory shared by users"""
    return PlatformDirs().site_bin_dir


def user_applications_dir() -> str:
    """:returns: applications directory tied to the user"""
    return PlatformDirs().user_applications_dir


def site_applications_dir(
    multipath: bool = False,  # noqa: FBT001, FBT002
    ensure_exists: bool = False,  # noqa: FBT001, FBT002
) -> str:
    """:param multipath: See `multipath <platformdirs.api.PlatformDirsABC.multipath>`.
    :param ensure_exists: See `ensure_exists <platformdirs.api.PlatformDirsABC.ensure_exists>`.

    :returns: applications directory shared by users

    """
    return PlatformDirs(
        multipath=multipath,
        ensure_exists=ensure_exists,
    ).site_applications_dir


def user_runtime_dir(  # noqa: PLR0913, PLR0917
    appname: str | None = None,
    appauthor: str | Literal[False] | None = None,
    version: str | None = None,
    opinion: bool = True,  # noqa: FBT001, FBT002
    ensure_exists: bool = False,  # noqa: FBT001, FBT002
    use_site_for_root: bool = False,  # noqa: FBT001, FBT002
) -> str:
    """:param appname: See `appname <platformdirs.api.PlatformDirsABC.appname>`.
    :param appauthor: See `appauthor <platformdirs.api.PlatformDirsABC.appauthor>`.
    :param version: See `version <platformdirs.api.PlatformDirsABC.version>`.
    :param opinion: See `opinion <platformdirs.api.PlatformDirsABC.opinion>`.
    :param ensure_exists: See `ensure_exists <platformdirs.api.PlatformDirsABC.ensure_exists>`.
    :param use_site_for_root: See `use_site_for_root <platformdirs.api.PlatformDirsABC.use_site_for_root>`.

    :returns: runtime directory tied to the user

    """
    return PlatformDirs(
        appname=appname,
        appauthor=appauthor,
        version=version,
        opinion=opinion,
        ensure_exists=ensure_exists,
        use_site_for_root=use_site_for_root,
    ).user_runtime_dir


def site_runtime_dir(
    appname: str | None = None,
    appauthor: str | Literal[False] | None = None,
    version: str | None = None,
    opinion: bool = True,  # noqa: FBT001, FBT002
    ensure_exists: bool = False,  # noqa: FBT001, FBT002
) -> str:
    """:param appname: See `appname <platformdirs.api.PlatformDirsABC.appname>`.
    :param appauthor: See `appauthor <platformdirs.api.PlatformDirsABC.appauthor>`.
    :param version: See `version <platformdirs.api.PlatformDirsABC.version>`.
    :param opinion: See `opinion <platformdirs.api.PlatformDirsABC.opinion>`.
    :param ensure_exists: See `ensure_exists <platformdirs.api.PlatformDirsABC.ensure_exists>`.

    :returns: runtime directory shared by users

    """
    return PlatformDirs(
        appname=appname,
        appauthor=appauthor,
        version=version,
        opinion=opinion,
        ensure_exists=ensure_exists,
    ).site_runtime_dir


def user_data_path(  # noqa: PLR0913, PLR0917
    appname: str | None = None,
    appauthor: str | Literal[False] | None = None,
    version: str | None = None,
    roaming: bool = False,  # noqa: FBT001, FBT002
    ensure_exists: bool = False,  # noqa: FBT001, FBT002
    use_site_for_root: bool = False,  # noqa: FBT001, FBT002
) -> Path:
    """:param appname: See `appname <platformdirs.api.PlatformDirsABC.appname>`.
    :param appauthor: See `appauthor <platformdirs.api.PlatformDirsABC.appauthor>`.
    :param version: See `version <platformdirs.api.PlatformDirsABC.version>`.
    :param roaming: See `roaming <platformdirs.api.PlatformDirsABC.roaming>`.
    :param ensure_exists: See `ensure_exists <platformdirs.api.PlatformDirsABC.ensure_exists>`.
    :param use_site_for_root: See `use_site_for_root <platformdirs.api.PlatformDirsABC.use_site_for_root>`.

    :returns: data path tied to the user

    """
    return PlatformDirs(
        appname=appname,
        appauthor=appauthor,
        version=version,
        roaming=roaming,
        ensure_exists=ensure_exists,
        use_site_for_root=use_site_for_root,
    ).user_data_path


def site_data_path(
    appname: str | None = None,
    appauthor: str | Literal[False] | None = None,
    version: str | None = None,
    multipath: bool = False,  # noqa: FBT001, FBT002
    ensure_exists: bool = False,  # noqa: FBT001, FBT002
) -> Path:
    """:param appname: See `appname <platformdirs.api.PlatformDirsABC.appname>`.
    :param appauthor: See `appauthor <platformdirs.api.PlatformDirsABC.appauthor>`.
    :param version: See `version <platformdirs.api.PlatformDirsABC.version>`.
    :param multipath: See `multipath <platformdirs.api.PlatformDirsABC.multipath>`.
    :param ensure_exists: See `ensure_exists <platformdirs.api.PlatformDirsABC.ensure_exists>`.

    :returns: data path shared by users

    """
    return PlatformDirs(
        appname=appname,
        appauthor=appauthor,
        version=version,
        multipath=multipath,
        ensure_exists=ensure_exists,
    ).site_data_path


def user_config_path(  # noqa: PLR0913, PLR0917
    appname: str | None = None,
    appauthor: str | Literal[False] | None = None,
    version: str | None = None,
    roaming: bool = False,  # noqa: FBT001, FBT002
    ensure_exists: bool = False,  # noqa: FBT001, FBT002
    use_site_for_root: bool = False,  # noqa: FBT001, FBT002
) -> Path:
    """:param appname: See `appname <platformdirs.api.PlatformDirsABC.appname>`.
    :param appauthor: See `appauthor <platformdirs.api.PlatformDirsABC.appauthor>`.
    :param version: See `version <platformdirs.api.PlatformDirsABC.version>`.
    :param roaming: See `roaming <platformdirs.api.PlatformDirsABC.roaming>`.
    :param ensure_exists: See `ensure_exists <platformdirs.api.PlatformDirsABC.ensure_exists>`.
    :param use_site_for_root: See `use_site_for_root <platformdirs.api.PlatformDirsABC.use_site_for_root>`.

    :returns: config path tied to the user

    """
    return PlatformDirs(
        appname=appname,
        appauthor=appauthor,
        version=version,
        roaming=roaming,
        ensure_exists=ensure_exists,
        use_site_for_root=use_site_for_root,
    ).user_config_path


def site_config_path(
    appname: str | None = None,
    appauthor: str | Literal[False] | None = None,
    version: str | None = None,
    multipath: bool = False,  # noqa: FBT001, FBT002
    ensure_exists: bool = False,  # noqa: FBT001, FBT002
) -> Path:
    """:param appname: See `appname <platformdirs.api.PlatformDirsABC.appname>`.
    :param appauthor: See `appauthor <platformdirs.api.PlatformDirsABC.appauthor>`.
    :param version: See `version <platformdirs.api.PlatformDirsABC.version>`.
    :param multipath: See `multipath <platformdirs.api.PlatformDirsABC.multipath>`.
    :param ensure_exists: See `ensure_exists <platformdirs.api.PlatformDirsABC.ensure_exists>`.

    :returns: config path shared by users

    """
    return PlatformDirs(
        appname=appname,
        appauthor=appauthor,
        version=version,
        multipath=multipath,
        ensure_exists=ensure_exists,
    ).site_config_path


def site_cache_path(
    appname: str | None = None,
    appauthor: str | Literal[False] | None = None,
    version: str | None = None,
    opinion: bool = True,  # noqa: FBT001, FBT002
    ensure_exists: bool = False,  # noqa: FBT001, FBT002
) -> Path:
    """:param appname: See `appname <platformdirs.api.PlatformDirsABC.appname>`.
    :param appauthor: See `appauthor <platformdirs.api.PlatformDirsABC.appauthor>`.
    :param version: See `version <platformdirs.api.PlatformDirsABC.version>`.
    :param opinion: See `opinion <platformdirs.api.PlatformDirsABC.opinion>`.
    :param ensure_exists: See `ensure_exists <platformdirs.api.PlatformDirsABC.ensure_exists>`.

    :returns: cache path shared by users

    """
    return PlatformDirs(
        appname=appname,
        appauthor=appauthor,
        version=version,
        opinion=opinion,
        ensure_exists=ensure_exists,
    ).site_cache_path


def user_cache_path(  # noqa: PLR0913, PLR0917
    appname: str | None = None,
    appauthor: str | Literal[False] | None = None,
    version: str | None = None,
    opinion: bool = True,  # noqa: FBT001, FBT002
    ensure_exists: bool = False,  # noqa: FBT001, FBT002
    use_site_for_root: bool = False,  # noqa: FBT001, FBT002
) -> Path:
    """:param appname: See `appname <platformdirs.api.PlatformDirsABC.appname>`.
    :param appauthor: See `appauthor <platformdirs.api.PlatformDirsABC.appauthor>`.
    :param version: See `version <platformdirs.api.PlatformDirsABC.version>`.
    :param opinion: See `opinion <platformdirs.api.PlatformDirsABC.opinion>`.
    :param ensure_exists: See `ensure_exists <platformdirs.api.PlatformDirsABC.ensure_exists>`.
    :param use_site_for_root: See `use_site_for_root <platformdirs.api.PlatformDirsABC.use_site_for_root>`.

    :returns: cache path tied to the user

    """
    return PlatformDirs(
        appname=appname,
        appauthor=appauthor,
        version=version,
        opinion=opinion,
        ensure_exists=ensure_exists,
        use_site_for_root=use_site_for_root,
    ).user_cache_path


def user_state_path(  # noqa: PLR0913, PLR0917
    appname: str | None = None,
    appauthor: str | Literal[False] | None = None,
    version: str | None = None,
    roaming: bool = False,  # noqa: FBT001, FBT002
    ensure_exists: bool = False,  # noqa: FBT001, FBT002
    use_site_for_root: bool = False,  # noqa: FBT001, FBT002
) -> Path:
    """:param appname: See `appname <platformdirs.api.PlatformDirsABC.appname>`.
    :param appauthor: See `appauthor <platformdirs.api.PlatformDirsABC.appauthor>`.
    :param version: See `version <platformdirs.api.PlatformDirsABC.version>`.
    :param roaming: See `roaming <platformdirs.api.PlatformDirsABC.roaming>`.
    :param ensure_exists: See `ensure_exists <platformdirs.api.PlatformDirsABC.ensure_exists>`.
    :param use_site_for_root: See `use_site_for_root <platformdirs.api.PlatformDirsABC.use_site_for_root>`.

    :returns: state path tied to the user

    """
    return PlatformDirs(
        appname=appname,
        appauthor=appauthor,
        version=version,
        roaming=roaming,
        ensure_exists=ensure_exists,
        use_site_for_root=use_site_for_root,
    ).user_state_path


def site_state_path(
    appname: str | None = None,
    appauthor: str | Literal[False] | None = None,
    version: str | None = None,
    ensure_exists: bool = False,  # noqa: FBT001, FBT002
) -> Path:
    """:param appname: See `appname <platformdirs.api.PlatformDirsABC.appname>`.
    :param appauthor: See `appauthor <platformdirs.api.PlatformDirsABC.appauthor>`.
    :param version: See `version <platformdirs.api.PlatformDirsABC.version>`.
    :param ensure_exists: See `ensure_exists <platformdirs.api.PlatformDirsABC.ensure_exists>`.

    :returns: state path shared by users

    """
    return PlatformDirs(
        appname=appname,
        appauthor=appauthor,
        version=version,
        ensure_exists=ensure_exists,
    ).site_state_path


def user_log_path(  # noqa: PLR0913, PLR0917
    appname: str | None = None,
    appauthor: str | Literal[False] | None = None,
    version: str | None = None,
    opinion: bool = True,  # noqa: FBT001, FBT002
    ensure_exists: bool = False,  # noqa: FBT001, FBT002
    use_site_for_root: bool = False,  # noqa: FBT001, FBT002
) -> Path:
    """:param appname: See `appname <platformdirs.api.PlatformDirsABC.appname>`.
    :param appauthor: See `appauthor <platformdirs.api.PlatformDirsABC.appauthor>`.
    :param version: See `version <platformdirs.api.PlatformDirsABC.version>`.
    :param opinion: See `opinion <platformdirs.api.PlatformDirsABC.opinion>`.
    :param ensure_exists: See `ensure_exists <platformdirs.api.PlatformDirsABC.ensure_exists>`.
    :param use_site_for_root: See `use_site_for_root <platformdirs.api.PlatformDirsABC.use_site_for_root>`.

    :returns: log path tied to the user

    """
    return PlatformDirs(
        appname=appname,
        appauthor=appauthor,
        version=version,
        opinion=opinion,
        ensure_exists=ensure_exists,
        use_site_for_root=use_site_for_root,
    ).user_log_path


def site_log_path(
    appname: str | None = None,
    appauthor: str | Literal[False] | None = None,
    version: str | None = None,
    opinion: bool = True,  # noqa: FBT001, FBT002
    ensure_exists: bool = False,  # noqa: FBT001, FBT002
) -> Path:
    """:param appname: See `appname <platformdirs.api.PlatformDirsABC.appname>`.
    :param appauthor: See `appauthor <platformdirs.api.PlatformDirsABC.appauthor>`.
    :param version: See `version <platformdirs.api.PlatformDirsABC.version>`.
    :param opinion: See `opinion <platformdirs.api.PlatformDirsABC.opinion>`.
    :param ensure_exists: See `ensure_exists <platformdirs.api.PlatformDirsABC.ensure_exists>`.

    :returns: log path shared by users

    """
    return PlatformDirs(
        appname=appname,
        appauthor=appauthor,
        version=version,
        opinion=opinion,
        ensure_exists=ensure_exists,
    ).site_log_path


def user_documents_path() -> Path:
    """:returns: documents path tied to the user"""
    return PlatformDirs().user_documents_path


def user_downloads_path() -> Path:
    """:returns: downloads path tied to the user"""
    return PlatformDirs().user_downloads_path


def user_pictures_path() -> Path:
    """:returns: pictures path tied to the user"""
    return PlatformDirs().user_pictures_path


def user_videos_path() -> Path:
    """:returns: videos path tied to the user"""
    return PlatformDirs().user_videos_path


def user_music_path() -> Path:
    """:returns: music path tied to the user"""
    return PlatformDirs().user_music_path


def user_desktop_path() -> Path:
    """:returns: desktop path tied to the user"""
    return PlatformDirs().user_desktop_path


def user_projects_path() -> Path:
    """:returns: projects path tied to the user"""
    return PlatformDirs().user_projects_path


def user_publicshare_path() -> Path:
    """:returns: public share path tied to the user"""
    return PlatformDirs().user_publicshare_path


def user_templates_path() -> Path:
    """:returns: templates path tied to the user"""
    return PlatformDirs().user_templates_path


def user_fonts_path() -> Path:
    """:returns: fonts path tied to the user"""
    return PlatformDirs().user_fonts_path


def user_preference_path() -> Path:
    """:returns: preference path tied to the user"""
    return PlatformDirs().user_preference_path


def user_bin_path() -> Path:
    """:returns: bin path tied to the user"""
    return PlatformDirs().user_bin_path


def site_bin_path() -> Path:
    """:returns: bin path shared by users"""
    return PlatformDirs().site_bin_path


def user_applications_path() -> Path:
    """:returns: applications path tied to the user"""
    return PlatformDirs().user_applications_path


def site_applications_path(
    multipath: bool = False,  # noqa: FBT001, FBT002
    ensure_exists: bool = False,  # noqa: FBT001, FBT002
) -> Path:
    """:param multipath: See `multipath <platformdirs.api.PlatformDirsABC.multipath>`.
    :param ensure_exists: See `ensure_exists <platformdirs.api.PlatformDirsABC.ensure_exists>`.

    :returns: applications path shared by users

    """
    return PlatformDirs(
        multipath=multipath,
        ensure_exists=ensure_exists,
    ).site_applications_path


def user_runtime_path(  # noqa: PLR0913, PLR0917
    appname: str | None = None,
    appauthor: str | Literal[False] | None = None,
    version: str | None = None,
    opinion: bool = True,  # noqa: FBT001, FBT002
    ensure_exists: bool = False,  # noqa: FBT001, FBT002
    use_site_for_root: bool = False,  # noqa: FBT001, FBT002
) -> Path:
    """:param appname: See `appname <platformdirs.api.PlatformDirsABC.appname>`.
    :param appauthor: See `appauthor <platformdirs.api.PlatformDirsABC.appauthor>`.
    :param version: See `version <platformdirs.api.PlatformDirsABC.version>`.
    :param opinion: See `opinion <platformdirs.api.PlatformDirsABC.opinion>`.
    :param ensure_exists: See `ensure_exists <platformdirs.api.PlatformDirsABC.ensure_exists>`.
    :param use_site_for_root: See `use_site_for_root <platformdirs.api.PlatformDirsABC.use_site_for_root>`.

    :returns: runtime path tied to the user

    """
    return PlatformDirs(
        appname=appname,
        appauthor=appauthor,
        version=version,
        opinion=opinion,
        ensure_exists=ensure_exists,
        use_site_for_root=use_site_for_root,
    ).user_runtime_path


def site_runtime_path(
    appname: str | None = None,
    appauthor: str | Literal[False] | None = None,
    version: str | None = None,
    opinion: bool = True,  # noqa: FBT001, FBT002
    ensure_exists: bool = False,  # noqa: FBT001, FBT002
) -> Path:
    """:param appname: See `appname <platformdirs.api.PlatformDirsABC.appname>`.
    :param appauthor: See `appauthor <platformdirs.api.PlatformDirsABC.appauthor>`.
    :param version: See `version <platformdirs.api.PlatformDirsABC.version>`.
    :param opinion: See `opinion <platformdirs.api.PlatformDirsABC.opinion>`.
    :param ensure_exists: See `ensure_exists <platformdirs.api.PlatformDirsABC.ensure_exists>`.

    :returns: runtime path shared by users

    """
    return PlatformDirs(
        appname=appname,
        appauthor=appauthor,
        version=version,
        opinion=opinion,
        ensure_exists=ensure_exists,
    ).site_runtime_path


__all__ = [
    "AppDirs",
    "PlatformDirs",
    "PlatformDirsABC",
    "__version__",
    "__version_info__",
    "site_applications_dir",
    "site_applications_path",
    "site_bin_dir",
    "site_bin_path",
    "site_cache_dir",
    "site_cache_path",
    "site_config_dir",
    "site_config_path",
    "site_data_dir",
    "site_data_path",
    "site_log_dir",
    "site_log_path",
    "site_runtime_dir",
    "site_runtime_path",
    "site_state_dir",
    "site_state_path",
    "user_applications_dir",
    "user_applications_path",
    "user_bin_dir",
    "user_bin_path",
    "user_cache_dir",
    "user_cache_path",
    "user_config_dir",
    "user_config_path",
    "user_data_dir",
    "user_data_path",
    "user_desktop_dir",
    "user_desktop_path",
    "user_documents_dir",
    "user_documents_path",
    "user_downloads_dir",
    "user_downloads_path",
    "user_fonts_dir",
    "user_fonts_path",
    "user_log_dir",
    "user_log_path",
    "user_music_dir",
    "user_music_path",
    "user_pictures_dir",
    "user_pictures_path",
    "user_preference_dir",
    "user_preference_path",
    "user_projects_dir",
    "user_projects_path",
    "user_publicshare_dir",
    "user_publicshare_path",
    "user_runtime_dir",
    "user_runtime_path",
    "user_state_dir",
    "user_state_path",
    "user_templates_dir",
    "user_templates_path",
    "user_videos_dir",
    "user_videos_path",
]



from __future__ import annotations

import os
import typing as t
from pathlib import Path

import platformdirs

from findpython.providers.rye import RyeProvider


class UvProvider(RyeProvider):
    @classmethod
    def create(cls) -> t.Self | None:
        default_root_str = platformdirs.user_data_dir("uv", appauthor=False, roaming=True)
        root_str = os.getenv("UV_PYTHON_INSTALL_DIR")
        if root_str is None:
            root = Path(default_root_str).expanduser() / "python"
        else:
            root = Path(root_str).expanduser()
        return cls(root)



# This file is dual licensed under the terms of the Apache License, Version
# 2.0, and the BSD License. See the LICENSE file in the root of this repository
# for complete details.

import typing

from cryptography.hazmat.primitives import hashes
from cryptography.utils import Buffer

class Hash(hashes.HashContext):
    def __init__(
        self, algorithm: hashes.HashAlgorithm, backend: typing.Any = None
    ) -> None: ...
    @property
    def algorithm(self) -> hashes.HashAlgorithm: ...
    def update(self, data: Buffer) -> None: ...
    def finalize(self) -> bytes: ...
    def copy(self) -> Hash: ...
    @staticmethod
    def hash(algorithm: hashes.HashAlgorithm, data: Buffer) -> bytes: ...

def hash_supported(algorithm: hashes.HashAlgorithm) -> bool: ...

class XOFHash:
    def __init__(self, algorithm: hashes.ExtendableOutputFunction) -> None: ...
    @property
    def algorithm(self) -> hashes.ExtendableOutputFunction: ...
    def update(self, data: Buffer) -> None: ...
    def squeeze(self, length: int) -> bytes: ...
    def copy(self) -> XOFHash: ...



from __future__ import annotations

import re
from copy import deepcopy
from functools import cached_property
from itertools import product
from os import environ
from typing import TYPE_CHECKING, Any

from hatch.env.utils import ensure_valid_environment
from hatch.project.constants import DEFAULT_BUILD_DIRECTORY, BuildEnvVars
from hatch.project.env import apply_overrides
from hatch.project.utils import format_script_commands, parse_script_command

if TYPE_CHECKING:
    from packaging.requirements import Requirement


class ProjectConfig:
    def __init__(self, root, config, plugin_manager=None):
        self.root = root
        self.config = config
        self.plugin_manager = plugin_manager

        self._matrices = None
        self._env = None
        self._env_requires_complex = None
        self._env_requires = None
        self._env_collectors = None
        self._envs = None
        self._internal_envs = None
        self._internal_matrices = None
        self._matrix_variables = None
        self._publish = None
        self._scripts = None
        self._cached_env_overrides = {}

    @cached_property
    def build(self):
        config = self.config.get('build', {})
        if not isinstance(config, dict):
            message = 'Field `tool.hatch.build` must be a table'
            raise TypeError(message)

        return BuildConfig(config)

    @property
    def env(self):
        if self._env is None:
            config = self.config.get('env', {})
            if not isinstance(config, dict):
                message = 'Field `tool.hatch.env` must be a table'
                raise TypeError(message)

            self._env = config

        return self._env

    @property
    def env_requires_complex(self) -> list[Requirement]:
        if self._env_requires_complex is None:
            from packaging.requirements import InvalidRequirement, Requirement

            requires = self.env.get('requires', [])
            if not isinstance(requires, list):
                message = 'Field `tool.hatch.env.requires` must be an array'
                raise TypeError(message)

            requires_complex = []

            for i, entry in enumerate(requires, 1):
                if not isinstance(entry, str):
                    message = f'Requirement #{i} in `tool.hatch.env.requires` must be a string'
                    raise TypeError(message)

                try:
                    requires_complex.append(Requirement(entry))
                except InvalidRequirement as e:
                    message = f'Requirement #{i} in `tool.hatch.env.requires` is invalid: {e}'
                    raise ValueError(message) from None

            self._env_requires_complex = requires_complex

        return self._env_requires_complex

    @property
    def env_requires(self):
        if self._env_requires is None:
            self._env_requires = [str(r) for r in self.env_requires_complex]

        return self._env_requires

    @property
    def env_collectors(self):
        if self._env_collectors is None:
            collectors = self.env.get('collectors', {})
            if not isinstance(collectors, dict):
                message = 'Field `tool.hatch.env.collectors` must be a table'
                raise TypeError(message)

            final_config = {'default': {}}
            for collector, config in collectors.items():
                if not isinstance(config, dict):
                    message = f'Field `tool.hatch.env.collectors.{collector}` must be a table'
                    raise TypeError(message)

                final_config[collector] = config

            self._env_collectors = final_config

        return self._env_collectors

    @property
    def matrices(self):
        if self._matrices is None:
            _ = self.envs

        return self._matrices

    @property
    def matrix_variables(self):
        if self._matrix_variables is None:
            _ = self.envs

        return self._matrix_variables

    @property
    def internal_envs(self):
        if self._internal_envs is None:
            _ = self.envs

        return self._internal_envs

    @property
    def internal_matrices(self):
        if self._internal_matrices is None:
            _ = self.envs

        return self._internal_matrices

    @property
    def envs(self):
        from hatch.env.internal import get_internal_env_config
        from hatch.utils.platform import get_platform_name

        if self._envs is None:
            env_config = self.config.get('envs', {})
            if not isinstance(env_config, dict):
                message = 'Field `tool.hatch.envs` must be a table'
                raise TypeError(message)

            config = {}
            environment_collectors = []

            for collector, collector_config in self.env_collectors.items():
                collector_class = self.plugin_manager.environment_collector.get(collector)
                if collector_class is None:
                    from hatchling.plugin.exceptions import UnknownPluginError

                    message = f'Unknown environment collector: {collector}'
                    raise UnknownPluginError(message)

                environment_collector = collector_class(self.root, collector_config)
                environment_collectors.append(environment_collector)

                for env_name, data in environment_collector.get_initial_config().items():
                    config.setdefault(env_name, data)

            for env_name, data in env_config.items():
                if not isinstance(data, dict):
                    message = f'Field `tool.hatch.envs.{env_name}` must be a table'
                    raise TypeError(message)

                config.setdefault(env_name, {}).update(data)

            for environment_collector in environment_collectors:
                environment_collector.finalize_config(config)

            # Prevent plugins from removing the default environment
            ensure_valid_environment(config.setdefault('default', {}))

            seen = set()
            active = []
            for env_name, data in config.items():
                _populate_default_env_values(env_name, data, config, seen, active)

            current_platform = get_platform_name()
            all_matrices = {}
            generated_envs = {}
            final_config = {}
            cached_overrides = {}
            for env_name, raw_initial_config in config.items():
                current_cached_overrides = cached_overrides[env_name] = {
                    'platform': [],
                    'env': [],
                    'matrix': [],
                    'name': [],
                }

                # Only shallow copying is necessary since we just want to modify keys
                initial_config = raw_initial_config.copy()

                matrix_name_format = initial_config.pop('matrix-name-format', '{value}')
                if not isinstance(matrix_name_format, str):
                    message = f'Field `tool.hatch.envs.{env_name}.matrix-name-format` must be a string'
                    raise TypeError(message)

                if '{value}' not in matrix_name_format:
                    message = (
                        f'Field `tool.hatch.envs.{env_name}.matrix-name-format` must '
                        f'contain at least the `{{value}}` placeholder'
                    )
                    raise ValueError(message)

                overrides = initial_config.pop('overrides', {})
                if not isinstance(overrides, dict):
                    message = f'Field `tool.hatch.envs.{env_name}.overrides` must be a table'
                    raise TypeError(message)

                # Apply any configuration based on the current platform
                platform_overrides = overrides.get('platform', {})
                if not isinstance(platform_overrides, dict):
                    message = f'Field `tool.hatch.envs.{env_name}.overrides.platform` must be a table'
                    raise TypeError(message)

                for platform, options in platform_overrides.items():
                    if not isinstance(options, dict):
                        message = f'Field `tool.hatch.envs.{env_name}.overrides.platform.{platform}` must be a table'
                        raise TypeError(message)

                    if platform != current_platform:
                        continue

                    apply_overrides(env_name, 'platform', platform, current_platform, options, initial_config)
                    current_cached_overrides['platform'].append((platform, current_platform, options))

                # Apply any configuration based on environment variables
                env_var_overrides = overrides.get('env', {})
                if not isinstance(env_var_overrides, dict):
                    message = f'Field `tool.hatch.envs.{env_name}.overrides.env` must be a table'
                    raise TypeError(message)

                for env_var, options in env_var_overrides.items():
                    if not isinstance(options, dict):
                        message = f'Field `tool.hatch.envs.{env_name}.overrides.env.{env_var}` must be a table'
                        raise TypeError(message)

                    if env_var not in environ:
                        continue

                    apply_overrides(env_name, 'env', env_var, environ[env_var], options, initial_config)
                    current_cached_overrides['env'].append((env_var, environ[env_var], options))

                if 'matrix' not in initial_config:
                    final_config[env_name] = initial_config
                    continue

                matrices = initial_config.pop('matrix')
                if not isinstance(matrices, list):
                    message = f'Field `tool.hatch.envs.{env_name}.matrix` must be an array'
                    raise TypeError(message)

                matrix_overrides = overrides.get('matrix', {})
                if not isinstance(matrix_overrides, dict):
                    message = f'Field `tool.hatch.envs.{env_name}.overrides.matrix` must be a table'
                    raise TypeError(message)

                name_overrides = overrides.get('name', {})
                if not isinstance(name_overrides, dict):
                    message = f'Field `tool.hatch.envs.{env_name}.overrides.name` must be a table'
                    raise TypeError(message)

                matrix_data = all_matrices[env_name] = {'config': deepcopy(initial_config)}
                all_envs = matrix_data['envs'] = {}
                for i, raw_matrix in enumerate(matrices, 1):
                    matrix = raw_matrix
                    if not isinstance(matrix, dict):
                        message = f'Entry #{i} in field `tool.hatch.envs.{env_name}.matrix` must be a table'
                        raise TypeError(message)

                    if not matrix:
                        message = f'Matrix #{i} in field `tool.hatch.envs.{env_name}.matrix` cannot be empty'
                        raise ValueError(message)

                    for j, (variable, values) in enumerate(matrix.items(), 1):
                        if not variable:
                            message = (
                                f'Variable #{j} in matrix #{i} in field `tool.hatch.envs.{env_name}.matrix` '
                                f'cannot be an empty string'
                            )
                            raise ValueError(message)

                        if not isinstance(values, list):
                            message = (
                                f'Variable `{variable}` in matrix #{i} in field `tool.hatch.envs.{env_name}.matrix` '
                                f'must be an array'
                            )
                            raise TypeError(message)

                        if not values:
                            message = (
                                f'Variable `{variable}` in matrix #{i} in field `tool.hatch.envs.{env_name}.matrix` '
                                f'cannot be empty'
                            )
                            raise ValueError(message)

                        existing_values = set()
                        for k, value in enumerate(values, 1):
                            if not isinstance(value, str):
                                message = (
                                    f'Value #{k} of variable `{variable}` in matrix #{i} in field '
                                    f'`tool.hatch.envs.{env_name}.matrix` must be a string'
                                )
                                raise TypeError(message)

                            if not value:
                                message = (
                                    f'Value #{k} of variable `{variable}` in matrix #{i} in field '
                                    f'`tool.hatch.envs.{env_name}.matrix` cannot be an empty string'
                                )
                                raise ValueError(message)

                            if value in existing_values:
                                message = (
                                    f'Value #{k} of variable `{variable}` in matrix #{i} in field '
                                    f'`tool.hatch.envs.{env_name}.matrix` is a duplicate'
                                )
                                raise ValueError(message)

                            existing_values.add(value)

                    variables = {}

                    # Ensure that any Python variable comes first
                    python_selected = False
                    for variable in ('py', 'python'):
                        if variable in matrix:
                            if python_selected:
                                message = (
                                    f'Matrix #{i} in field `tool.hatch.envs.{env_name}.matrix` '
                                    f'cannot contain both `py` and `python` variables'
                                )
                                raise ValueError(message)
                            python_selected = True

                            # Only shallow copying is necessary since we just want to remove a key
                            matrix = matrix.copy()
                            variables[variable] = matrix.pop(variable)

                    variables.update(matrix)

                    for result in product(*variables.values()):
                        # Make a value mapping for easy referencing
                        variable_values = dict(zip(variables, result))

                        # Create the environment's initial configuration
                        new_config = deepcopy(initial_config)

                        cached_matrix_overrides = []

                        # Apply any configuration based on matrix variables
                        for variable, options in matrix_overrides.items():
                            if not isinstance(options, dict):
                                message = (
                                    f'Field `tool.hatch.envs.{env_name}.overrides.matrix.{variable}` must be a table'
                                )
                                raise TypeError(message)

                            if variable not in variables:
                                continue

                            apply_overrides(
                                env_name, 'matrix', variable, variable_values[variable], options, new_config
                            )
                            cached_matrix_overrides.append((variable, variable_values[variable], options))

                        # Construct the environment name
                        final_matrix_name_format = new_config.pop('matrix-name-format', matrix_name_format)
                        env_name_parts = []
                        for j, (variable, value) in enumerate(variable_values.items()):
                            if j == 0 and python_selected:
                                new_config['python'] = value
                                env_name_parts.append(value if value.startswith('py') else f'py{value}')
                            else:
                                env_name_parts.append(final_matrix_name_format.format(variable=variable, value=value))

                        new_env_name = '-'.join(env_name_parts)

                        cached_name_overrides = []

                        # Apply any configuration based on the final name, minus the prefix for non-default environments
                        for pattern, options in name_overrides.items():
                            if not isinstance(options, dict):
                                message = f'Field `tool.hatch.envs.{env_name}.overrides.name.{pattern}` must be a table'
                                raise TypeError(message)

                            if not re.search(pattern, new_env_name):
                                continue

                            apply_overrides(env_name, 'name', pattern, new_env_name, options, new_config)
                            cached_name_overrides.append((pattern, new_env_name, options))

                        if env_name != 'default':
                            new_env_name = f'{env_name}.{new_env_name}'

                        # Save the generated environment
                        final_config[new_env_name] = new_config
                        cached_overrides[new_env_name] = {
                            'platform': current_cached_overrides['platform'],
                            'env': current_cached_overrides['env'],
                            'matrix': cached_matrix_overrides,
                            'name': cached_name_overrides,
                        }
                        all_envs[new_env_name] = variable_values
                        if 'py' in variable_values:
                            all_envs[new_env_name] = {'python': variable_values.pop('py'), **variable_values}

                # Remove the root matrix generator
                del cached_overrides[env_name]

                # Save the variables used to generate the environments
                generated_envs.update(all_envs)

            for environment_collector in environment_collectors:
                environment_collector.finalize_environments(final_config)

            self._matrices = all_matrices
            self._internal_matrices = {}
            self._envs = final_config
            self._matrix_variables = generated_envs
            self._cached_env_overrides.update(cached_overrides)

            # Extract the internal environments
            self._internal_envs = {}
            for internal_name in get_internal_env_config():
                try:
                    self._internal_envs[internal_name] = self._envs.pop(internal_name)
                # Matrix
                except KeyError:
                    self._internal_matrices[internal_name] = self._matrices.pop(internal_name)
                    for env_name in [env_name for env_name in self._envs if env_name.startswith(f'{internal_name}.')]:
                        self._internal_envs[env_name] = self._envs.pop(env_name)

        return self._envs

    @property
    def publish(self):
        if self._publish is None:
            config = self.config.get('publish', {})
            if not isinstance(config, dict):
                message = 'Field `tool.hatch.publish` must be a table'
                raise TypeError(message)

            for publisher, data in config.items():
                if not isinstance(data, dict):
                    message = f'Field `tool.hatch.publish.{publisher}` must be a table'
                    raise TypeError(message)

            self._publish = config

        return self._publish

    @property
    def scripts(self):
        if self._scripts is None:
            script_config = self.config.get('scripts', {})
            if not isinstance(script_config, dict):
                message = 'Field `tool.hatch.scripts` must be a table'
                raise TypeError(message)

            config = {}

            for name, data in script_config.items():
                if ' ' in name:
                    message = f'Script name `{name}` in field `tool.hatch.scripts` must not contain spaces'
                    raise ValueError(message)

                commands = []

                if isinstance(data, str):
                    commands.append(data)
                elif isinstance(data, list):
                    for i, command in enumerate(data, 1):
                        if not isinstance(command, str):
                            message = f'Command #{i} in field `tool.hatch.scripts.{name}` must be a string'
                            raise TypeError(message)

                        commands.append(command)
                else:
                    message = f'Field `tool.hatch.scripts.{name}` must be a string or an array of strings'
                    raise TypeError(message)

                config[name] = commands

            seen = {}
            active = []
            for script_name, commands in config.items():
                commands[:] = expand_script_commands(script_name, commands, config, seen, active)

            self._scripts = config

        return self._scripts

    def finalize_env_overrides(self, option_types):
        # We lazily apply overrides because we need type information potentially defined by
        # environment plugins for their options
        if not self._cached_env_overrides:
            return

        for environments in (self.envs, self.internal_envs):
            for env_name, config in environments.items():
                for override_name, data in self._cached_env_overrides.get(env_name, {}).items():
                    for condition, condition_value, options in data:
                        apply_overrides(
                            env_name, override_name, condition, condition_value, options, config, option_types
                        )

        self._cached_env_overrides.clear()


class BuildConfig:
    def __init__(self, config: dict[str, Any]) -> None:
        self.__config = config
        self.__targets: dict[str, BuildTargetConfig] = config

    @cached_property
    def directory(self) -> str:
        directory = self.__config.get('directory', DEFAULT_BUILD_DIRECTORY)
        if not isinstance(directory, str):
            message = 'Field `tool.hatch.build.directory` must be a string'
            raise TypeError(message)

        return directory

    @cached_property
    def dependencies(self) -> list[str]:
        dependencies: list[str] = self.__config.get('dependencies', [])
        if not isinstance(dependencies, list):
            message = 'Field `tool.hatch.build.dependencies` must be an array'
            raise TypeError(message)

        for i, dependency in enumerate(dependencies, 1):
            if not isinstance(dependency, str):
                message = f'Dependency #{i} in field `tool.hatch.build.dependencies` must be a string'
                raise TypeError(message)

        return list(dependencies)

    @cached_property
    def hook_config(self) -> dict[str, dict[str, Any]]:
        hook_config: dict[str, dict[str, Any]] = self.__config.get('hooks', {})
        if not isinstance(hook_config, dict):
            message = 'Field `tool.hatch.build.hooks` must be a table'
            raise TypeError(message)

        for hook_name, config in hook_config.items():
            if not isinstance(config, dict):
                message = f'Field `tool.hatch.build.hooks.{hook_name}` must be a table'
                raise TypeError(message)

        return finalize_hook_config(hook_config)

    @cached_property
    def __target_config(self) -> dict[str, Any]:
        config = self.__config.get('targets', {})
        if not isinstance(config, dict):
            message = 'Field `tool.hatch.build.targets` must be a table'
            raise TypeError(message)

        return config

    def target(self, target: str) -> BuildTargetConfig:
        if target in self.__targets:
            return self.__targets[target]

        config = self.__target_config.get(target, {})
        if not isinstance(config, dict):
            message = f'Field `tool.hatch.build.targets.{target}` must be a table'
            raise TypeError(message)

        target_config = BuildTargetConfig(target, config, self)
        self.__targets[target] = target_config
        return target_config


class BuildTargetConfig:
    def __init__(self, name: str, config: dict[str, Any], global_config: BuildConfig) -> None:
        self.__name = name
        self.__config = config
        self.__global_config = global_config

    @cached_property
    def directory(self) -> str:
        directory = self.__config.get('directory', self.__global_config.directory)
        if not isinstance(directory, str):
            message = f'Field `tool.hatch.build.targets.{self.__name}.directory` must be a string'
            raise TypeError(message)

        return directory

    @cached_property
    def dependencies(self) -> list[str]:
        dependencies: list[str] = self.__config.get('dependencies', [])
        if not isinstance(dependencies, list):
            message = f'Field `tool.hatch.build.targets.{self.__name}.dependencies` must be an array'
            raise TypeError(message)

        for i, dependency in enumerate(dependencies, 1):
            if not isinstance(dependency, str):
                message = (
                    f'Dependency #{i} in field `tool.hatch.build.targets.{self.__name}.dependencies` must be a string'
                )
                raise TypeError(message)

        all_dependencies = list(self.__global_config.dependencies)
        all_dependencies.extend(dependencies)
        return all_dependencies

    @cached_property
    def hook_config(self) -> dict[str, dict[str, Any]]:
        hook_config: dict[str, dict[str, Any]] = self.__config.get('hooks', {})
        if not isinstance(hook_config, dict):
            message = f'Field `tool.hatch.build.targets.{self.__name}.hooks` must be a table'
            raise TypeError(message)

        for hook_name, config in hook_config.items():
            if not isinstance(config, dict):
                message = f'Field `tool.hatch.build.targets.{self.__name}.hooks.{hook_name}` must be a table'
                raise TypeError(message)

        config = self.__global_config.hook_config.copy()
        config.update(hook_config)
        return finalize_hook_config(config)


def expand_script_commands(script_name, commands, config, seen, active):
    if script_name in seen:
        return seen[script_name]

    if script_name in active:
        active.append(script_name)

        message = f'Circular expansion detected for field `tool.hatch.scripts`: {" -> ".join(active)}'
        raise ValueError(message)

    active.append(script_name)

    expanded_commands = []

    for command in commands:
        possible_script, args, ignore_exit_code = parse_script_command(command)

        if possible_script in config:
            expanded_commands.extend(
                format_script_commands(
                    commands=expand_script_commands(possible_script, config[possible_script], config, seen, active),
                    args=args,
                    ignore_exit_code=ignore_exit_code,
                )
            )
        else:
            expanded_commands.append(command)

    seen[script_name] = expanded_commands
    active.pop()

    return expanded_commands


def _populate_default_env_values(env_name, data, config, seen, active):
    if env_name in seen:
        return

    if data.pop('detached', False):
        data['template'] = env_name
        data['skip-install'] = True

    template_name = data.pop('template', 'default')
    if template_name not in config:
        message = f'Field `tool.hatch.envs.{env_name}.template` refers to an unknown environment `{template_name}`'
        raise ValueError(message)

    if env_name in active:
        active.append(env_name)

        message = f'Circular inheritance detected for field `tool.hatch.envs.*.template`: {" -> ".join(active)}'
        raise ValueError(message)

    if template_name == env_name:
        ensure_valid_environment(data)
        seen.add(env_name)
        return

    active.append(env_name)

    template_config = config[template_name]
    _populate_default_env_values(template_name, template_config, config, seen, active)

    for key, value in template_config.items():
        if key == 'matrix':
            continue

        if key == 'scripts':
            scripts = data['scripts'] if 'scripts' in data else data.setdefault('scripts', {})
            for script, commands in value.items():
                scripts.setdefault(script, commands)
        else:
            data.setdefault(key, value)

    seen.add(env_name)
    active.pop()


def finalize_hook_config(hook_config: dict[str, dict[str, Any]]) -> dict[str, dict[str, Any]]:
    if env_var_enabled(BuildEnvVars.NO_HOOKS):
        return {}

    all_hooks_enabled = env_var_enabled(BuildEnvVars.HOOKS_ENABLE)
    final_hook_config: dict[str, dict[str, Any]] = {
        hook_name: config
        for hook_name, config in hook_config.items()
        if (
            all_hooks_enabled
            or config.get('enable-by-default', True)
            or env_var_enabled(f'{BuildEnvVars.HOOK_ENABLE_PREFIX}{hook_name.upper()}')
        )
    }
    return final_hook_config


def env_var_enabled(env_var: str, *, default: bool = False) -> bool:
    if env_var in environ:
        return environ[env_var] in {'1', 'true'}

    return default



# Author: David Goodger
# Contact: goodger@users.sourceforge.net
# Revision: $Revision: 2224 $
# Date: $Date: 2004-06-05 21:40:46 +0200 (Sat, 05 Jun 2004) $
# Copyright: This module has been placed in the public domain.

# New language mappings are welcome.  Before doing a new translation, please
# read <https://docutils.sourceforge.io/docs/howto/i18n.html>.
# Two files must be translated for each language: one in docutils/languages,
# the other in docutils/parsers/rst/languages.

"""
Galician-language mappings for language-dependent features of Docutils.
"""

__docformat__ = 'reStructuredText'

labels = {
      # fixed: language-dependent
      'author': 'Autor',
      'authors': 'Autores',
      'organization': 'Organización',
      'address': 'Enderezo',
      'contact': 'Contacto',
      'version': 'Versión',
      'revision': 'Revisión',
      'status': 'Estado',
      'date': 'Data',
      'copyright': 'Dereitos de copia',
      'dedication': 'Dedicatoria',
      'abstract': 'Abstract',
      'attention': 'Atención!',
      'caution': 'Advertencia!',
      'danger': 'PERIGO!',
      'error': 'Erro',
      'hint': 'Consello',
      'important': 'Importante',
      'note': 'Nota',
      'tip': 'Suxestión',
      'warning': 'Aviso',
      'contents': 'Contido'}
"""Mapping of node class name to label text."""

bibliographic_fields = {
      # language-dependent: fixed
      'autor': 'author',
      'autores': 'authors',
      'organización': 'organization',
      'enderezo': 'address',
      'contacto': 'contact',
      'versión': 'version',
      'revisión': 'revision',
      'estado': 'status',
      'data': 'date',
      'dereitos de copia': 'copyright',
      'dedicatoria': 'dedication',
      'abstract': 'abstract'}
"""Galician (lowcased) to canonical name mapping for bibliographic fields."""

author_separators = [';', ',']
"""List of separator strings for the 'Authors' bibliographic field. Tried in
order."""



from __future__ import annotations

import json
import sys
from functools import lru_cache
from typing import TYPE_CHECKING, Any, Literal

from hatch.utils.fs import Path
from hatch.utils.runner import ExecutionContext

if TYPE_CHECKING:
    from hatch.env.plugin.interface import EnvironmentInterface
    from hatch.project.core import Project


class BuildFrontend:
    def __init__(self, project: Project, env: EnvironmentInterface) -> None:
        self.__project = project
        self.__env = env
        self.__scripts = StandardBuildFrontendScripts(self.__project, self.__env)
        self.__hatch = HatchBuildFrontend(self.__project, self.__env)

    @property
    def scripts(self) -> StandardBuildFrontendScripts:
        return self.__scripts

    @property
    def hatch(self) -> HatchBuildFrontend:
        return self.__hatch

    def build_sdist(self, directory: Path) -> Path:
        with self.__env.fs_context() as fs_context:
            output_context = fs_context.join('output')
            output_context.local_path.ensure_dir_exists()
            script = self.scripts.build_sdist(project_root=self.__env.project_root, output_dir=output_context.env_path)

            script_context = fs_context.join('build_sdist.py')
            script_context.local_path.parent.ensure_dir_exists()
            script_context.local_path.write_text(script)
            script_context.sync_env()

            context = ExecutionContext(self.__env)
            context.add_shell_command(['python', '-u', script_context.env_path])
            self.__env.app.execute_context(context)
            output_context.sync_local()

            output_path = output_context.local_path / 'output.json'
            output = json.loads(output_path.read_text())

            work_dir = output_context.local_path / 'work'
            artifact_path = Path(work_dir / output['return_val'])
            artifact_path.move(directory)
            return directory / artifact_path.name

    def build_wheel(self, directory: Path) -> Path:
        with self.__env.fs_context() as fs_context:
            output_context = fs_context.join('output')
            output_context.local_path.ensure_dir_exists()
            script = self.scripts.build_wheel(project_root=self.__env.project_root, output_dir=output_context.env_path)

            script_context = fs_context.join('build_wheel.py')
            script_context.local_path.parent.ensure_dir_exists()
            script_context.local_path.write_text(script)
            script_context.sync_env()

            context = ExecutionContext(self.__env)
            context.add_shell_command(['python', '-u', script_context.env_path])
            self.__env.app.execute_context(context)
            output_context.sync_local()

            output_path = output_context.local_path / 'output.json'
            output = json.loads(output_path.read_text())

            work_dir = output_context.local_path / 'work'
            artifact_path = Path(work_dir / output['return_val'])
            artifact_path.move(directory)
            return directory / artifact_path.name

    def get_requires(self, build: Literal['sdist', 'wheel', 'editable']) -> list[str]:
        with self.__env.fs_context() as fs_context:
            output_context = fs_context.join('output')
            output_context.local_path.ensure_dir_exists()
            script = self.scripts.get_requires(
                project_root=self.__env.project_root, output_dir=output_context.env_path, build=build
            )

            script_context = fs_context.join(f'get_requires_{build}.py')
            script_context.local_path.parent.ensure_dir_exists()
            script_context.local_path.write_text(script)
            script_context.sync_env()

            context = ExecutionContext(self.__env)
            context.add_shell_command(['python', '-u', script_context.env_path])
            self.__env.app.execute_context(context)
            output_context.sync_local()

            output_path = output_context.local_path / 'output.json'
            output = json.loads(output_path.read_text())
            return output['return_val']

    def get_core_metadata(self, *, editable: bool = False) -> dict[str, Any]:
        from hatchling.metadata.spec import project_metadata_from_core_metadata

        with self.__env.fs_context() as fs_context:
            output_context = fs_context.join('output')
            output_context.local_path.ensure_dir_exists()
            script = self.scripts.prepare_metadata(
                project_root=self.__env.project_root, output_dir=output_context.env_path, editable=editable
            )

            script_context = fs_context.join('get_core_metadata.py')
            script_context.local_path.parent.ensure_dir_exists()
            script_context.local_path.write_text(script)
            script_context.sync_env()

            context = ExecutionContext(self.__env)
            context.add_shell_command(['python', '-u', script_context.env_path])
            self.__env.app.execute_context(context)
            output_context.sync_local()

            output_path = output_context.local_path / 'output.json'
            output = json.loads(output_path.read_text())

            work_dir = output_context.local_path / 'work'
            metadata_file = Path(work_dir) / output['return_val'] / 'METADATA'
            return project_metadata_from_core_metadata(metadata_file.read_text())


class HatchBuildFrontend:
    def __init__(self, project: Project, env: EnvironmentInterface) -> None:
        self.__project = project
        self.__env = env
        self.__scripts = HatchBuildFrontendScripts(self.__project, self.__env)

    @property
    def scripts(self) -> HatchBuildFrontendScripts:
        return self.__scripts

    def get_build_deps(self, targets: list[str]) -> list[str]:
        with self.__env.fs_context() as fs_context:
            output_context = fs_context.join('output')
            output_context.local_path.ensure_dir_exists()
            script = self.scripts.get_build_deps(
                project_root=self.__env.project_root, output_dir=output_context.env_path, targets=targets
            )

            script_context = fs_context.join(f'get_build_deps_{"_".join(targets)}.py')
            script_context.local_path.parent.ensure_dir_exists()
            script_context.local_path.write_text(script)
            script_context.sync_env()

            context = ExecutionContext(self.__env)
            context.add_shell_command(['python', '-u', script_context.env_path])
            self.__env.app.execute_context(context)
            output_context.sync_local()

            output_path = output_context.local_path / 'output.json'
            output: list[str] = json.loads(output_path.read_text())
            return output

    def get_core_metadata(self) -> dict[str, Any]:
        with self.__env.fs_context() as fs_context:
            output_context = fs_context.join('output')
            output_context.local_path.ensure_dir_exists()
            script = self.scripts.get_core_metadata(
                project_root=self.__env.project_root, output_dir=output_context.env_path
            )

            script_context = fs_context.join('get_core_metadata.py')
            script_context.local_path.parent.ensure_dir_exists()
            script_context.local_path.write_text(script)
            script_context.sync_env()

            context = ExecutionContext(self.__env)
            context.add_shell_command(['python', '-u', script_context.env_path])
            self.__env.app.execute_context(context)
            output_context.sync_local()

            output_path = output_context.local_path / 'output.json'
            output: dict[str, Any] = json.loads(output_path.read_text())
            return output

    def get_required_build_deps(self, targets: list[str]) -> list[str]:
        target_dependencies: list[str] = []
        hooks: set[str] = set()
        for target in targets:
            target_config = self.__project.config.build.target(target)
            target_dependencies.extend(target_config.dependencies)
            hooks.update(target_config.hook_config)

        # Remove any build hooks that are known to not define any dependencies dynamically
        hooks.difference_update((
            # Built-in
            'version',
            # Popular third-party
            'vcs',
        ))

        if hooks:
            return self.get_build_deps(targets)

        return target_dependencies


class BuildFrontendScripts:
    def __init__(self, project: Project, env: EnvironmentInterface) -> None:
        self._project = project
        self._env = env

    @staticmethod
    def inject_data(script: str, data: dict[str, Any]) -> str:
        # All scripts have a constant dictionary on top
        return script.replace('{}', repr(data), 1)


class StandardBuildFrontendScripts(BuildFrontendScripts):
    def get_runner_script(
        self,
        *,
        project_root: str,
        output_dir: str,
        hook: str,
        kwargs: dict[str, Any],
    ) -> str:
        return self.inject_data(
            runner_script(),
            {
                'project_root': project_root,
                'output_dir': output_dir,
                'hook': hook,
                'kwargs': kwargs,
                'backend': self._project.metadata.build.build_backend,
                'backend_path': self._env.pathsep.join(self._project.metadata.build.backend_path),
                'hook_caller_script': hook_caller_script(),
            },
        )

    def get_requires(
        self,
        *,
        project_root: str,
        output_dir: str,
        build: Literal['sdist', 'wheel', 'editable'],
    ) -> str:
        return self.get_runner_script(
            project_root=project_root,
            output_dir=output_dir,
            hook=f'get_requires_for_build_{build}',
            kwargs={'config_settings': None},
        )

    def prepare_metadata(self, *, output_dir: str, project_root: str, editable: bool = False) -> str:
        return self.get_runner_script(
            project_root=project_root,
            output_dir=output_dir,
            hook='prepare_metadata_for_build_editable' if editable else 'prepare_metadata_for_build_wheel',
            kwargs={'work_dir': 'metadata_directory', 'config_settings': None, '_allow_fallback': True},
        )

    def build_wheel(self, *, output_dir: str, project_root: str, editable: bool = False) -> str:
        return self.get_runner_script(
            project_root=project_root,
            output_dir=output_dir,
            hook='build_editable' if editable else 'build_wheel',
            kwargs={'work_dir': 'wheel_directory', 'config_settings': None, 'metadata_directory': None},
        )

    def build_sdist(self, *, output_dir: str, project_root: str) -> str:
        return self.get_runner_script(
            project_root=project_root,
            output_dir=output_dir,
            hook='build_sdist',
            kwargs={'work_dir': 'sdist_directory', 'config_settings': None},
        )


class HatchBuildFrontendScripts(BuildFrontendScripts):
    def get_build_deps(self, *, output_dir: str, project_root: str, targets: list[str]) -> str:
        return self.inject_data(
            hatch_build_deps_script(),
            {
                'project_root': project_root,
                'output_dir': output_dir,
                'targets': targets,
            },
        )

    def get_core_metadata(self, *, output_dir: str, project_root: str) -> str:
        return self.inject_data(
            hatch_core_metadata_script(),
            {
                'project_root': project_root,
                'output_dir': output_dir,
            },
        )


if sys.version_info[:2] >= (3, 9):

    @lru_cache(maxsize=None)
    def hook_caller_script() -> str:
        from importlib.resources import files

        script = files('pyproject_hooks._in_process') / '_in_process.py'
        return script.read_text(encoding='utf-8')

    @lru_cache(maxsize=None)
    def runner_script() -> str:
        from importlib.resources import files

        script = files('hatch.project.frontend.scripts') / 'standard.py'
        return script.read_text(encoding='utf-8')

    @lru_cache(maxsize=None)
    def hatch_build_deps_script() -> str:
        from importlib.resources import files

        script = files('hatch.project.frontend.scripts') / 'build_deps.py'
        return script.read_text(encoding='utf-8')

    @lru_cache(maxsize=None)
    def hatch_core_metadata_script() -> str:
        from importlib.resources import files

        script = files('hatch.project.frontend.scripts') / 'core_metadata.py'
        return script.read_text(encoding='utf-8')

else:

    @lru_cache(maxsize=None)
    def hook_caller_script() -> str:
        from importlib.resources import read_text

        return read_text('pyproject_hooks._in_process', '_in_process.py')

    @lru_cache(maxsize=None)
    def runner_script() -> str:
        from importlib.resources import read_text

        return read_text('hatch.project.frontend.scripts', 'standard.py')

    @lru_cache(maxsize=None)
    def hatch_build_deps_script() -> str:
        from importlib.resources import read_text

        return read_text('hatch.project.frontend.scripts', 'build_deps.py')

    @lru_cache(maxsize=None)
    def hatch_core_metadata_script() -> str:
        from importlib.resources import read_text

        return read_text('hatch.project.frontend.scripts', 'core_metadata.py')



import click


@click.command(short_help='Remove build artifacts')
@click.argument('location', required=False)
@click.option(
    '--target',
    '-t',
    'targets',
    multiple=True,
    help=(
        'The target with which to remove artifacts, overriding project defaults. '
        'This may be selected multiple times e.g. `-t sdist -t wheel`'
    ),
)
@click.option(
    '--hooks-only',
    is_flag=True,
    help='Whether or not to only remove artifacts from build hooks [env var: `HATCH_BUILD_HOOKS_ONLY`]',
)
@click.option(
    '--no-hooks',
    is_flag=True,
    help='Whether or not to ignore artifacts from build hooks [env var: `HATCH_BUILD_NO_HOOKS`]',
)
@click.option(
    '--ext',
    is_flag=True,
    help=(
        'Whether or not to only remove artifacts from build hooks for distributing binary Python packages, such as '
        'compiled extensions. Equivalent to `--hooks-only -t wheel`'
    ),
)
@click.pass_context
def clean(ctx, location, targets, hooks_only, no_hooks, ext):
    """Remove build artifacts."""
    from hatch.cli.build import build

    ctx.invoke(
        build, clean_only=True, location=location, targets=targets, hooks_only=hooks_only, no_hooks=no_hooks, ext=ext
    )



import click

from hatch.config.constants import PublishEnvVars


@click.command(short_help='Publish build artifacts')
@click.argument('artifacts', nargs=-1)
@click.option(
    '--repo',
    '-r',
    envvar=PublishEnvVars.REPO,
    help='The repository with which to publish artifacts [env var: `HATCH_INDEX_REPO`]',
)
@click.option(
    '--user', '-u', envvar=PublishEnvVars.USER, help='The user with which to authenticate [env var: `HATCH_INDEX_USER`]'
)
@click.option(
    '--auth',
    '-a',
    envvar=PublishEnvVars.AUTH,
    help='The credentials to use for authentication [env var: `HATCH_INDEX_AUTH`]',
)
@click.option(
    '--ca-cert',
    envvar=PublishEnvVars.CA_CERT,
    help='The path to a CA bundle [env var: `HATCH_INDEX_CA_CERT`]',
)
@click.option(
    '--client-cert',
    envvar=PublishEnvVars.CLIENT_CERT,
    help='The path to a client certificate, optionally containing the private key [env var: `HATCH_INDEX_CLIENT_CERT`]',
)
@click.option(
    '--client-key',
    envvar=PublishEnvVars.CLIENT_KEY,
    help="The path to the client certificate's private key [env var: `HATCH_INDEX_CLIENT_KEY`]",
)
@click.option('--no-prompt', '-n', is_flag=True, help='Disable prompts, such as for missing required fields')
@click.option(
    '--initialize-auth', is_flag=True, help='Save first-time authentication information even if nothing was published'
)
@click.option(
    '--publisher',
    '-p',
    'publisher_name',
    envvar=PublishEnvVars.PUBLISHER,
    default='index',
    help='The publisher plugin to use (default is `index`) [env var: `HATCH_PUBLISHER`]',
)
@click.option(
    '--option',
    '-o',
    'options',
    envvar=PublishEnvVars.OPTIONS,
    multiple=True,
    help=(
        'Options to pass to the publisher plugin. This may be selected multiple '
        'times e.g. `-o foo=bar -o baz=23` [env var: `HATCH_PUBLISHER_OPTIONS`]'
    ),
)
@click.option('--yes', '-y', is_flag=True, help='Confirm without prompting when the plugin is disabled')
@click.pass_obj
def publish(
    app,
    artifacts,
    repo,
    user,
    auth,
    ca_cert,
    client_cert,
    client_key,
    no_prompt,
    initialize_auth,
    publisher_name,
    options,
    yes,
):
    """Publish build artifacts."""
    option_map = {'no_prompt': no_prompt, 'initialize_auth': initialize_auth}
    if publisher_name == 'index':
        if options:
            app.abort('Use the standard CLI flags rather than passing explicit options when using the `index` plugin')

        if repo:
            option_map['repo'] = repo
        if user:
            option_map['user'] = user
        if auth:
            option_map['auth'] = auth
        if ca_cert:
            option_map['ca_cert'] = ca_cert
        if client_cert:
            option_map['client_cert'] = client_cert
        if client_key:
            option_map['client_key'] = client_key
    else:  # no cov
        for option in options:
            key, _, value = option.partition('=')
            option_map[key] = value

    publisher_class = app.plugins.publisher.get(publisher_name)
    if publisher_class is None:
        app.abort(f'Unknown publisher: {publisher_name}')

    publisher = publisher_class(
        app,
        app.project.location,
        app.cache_dir / 'publish' / publisher_name,
        app.project.config.publish.get(publisher_name, {}),
        app.config.publish.get(publisher_name, {}),
    )
    if publisher.disable and not (yes or (not no_prompt and app.confirm(f'Confirm `{publisher_name}` publishing'))):
        app.abort(f'Publisher is disabled: {publisher_name}')

    publisher.publish(list(artifacts), option_map)
